# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test tools for testing rounds."""

import re
from copy import deepcopy
from dataclasses import dataclass
from enum import Enum
from typing import (
    Any,
    Callable,
    FrozenSet,
    Generator,
    List,
    Mapping,
    Optional,
    Tuple,
    Type,
)
from unittest import mock

import pytest

from packages.valory.skills.abstract_round_abci.base import (
    ABCIAppInternalError,
    AbciAppDB,
    AbstractRound,
    BaseSynchronizedData,
    BaseTxPayload,
    CollectDifferentUntilAllRound,
    CollectDifferentUntilThresholdRound,
    CollectNonEmptyUntilThresholdRound,
    CollectSameUntilAllRound,
    CollectSameUntilThresholdRound,
    CollectionRound,
    OnlyKeeperSendsRound,
    TransactionNotValidError,
    VotingRound,
)


MAX_PARTICIPANTS: int = 4


def get_participants() -> FrozenSet[str]:
    """Participants"""
    return frozenset([f"agent_{i}" for i in range(MAX_PARTICIPANTS)])


class DummyEvent(Enum):
    """Dummy Event"""

    DONE = "done"
    ROUND_TIMEOUT = "round_timeout"
    NO_MAJORITY = "no_majority"
    RESET_TIMEOUT = "reset_timeout"
    NEGATIVE = "negative"
    NONE = "none"
    FAIL = "fail"


@dataclass(frozen=True)
class DummyTxPayload(BaseTxPayload):
    """Dummy Transaction Payload."""

    value: Optional[str] = None
    vote: Optional[bool] = None


class DummySynchronizedData(BaseSynchronizedData):
    """Dummy synchronized data for tests."""


def get_dummy_tx_payloads(
    participants: FrozenSet[str],
    value: Any = None,
    vote: Optional[bool] = False,
    is_value_none: bool = False,
    is_vote_none: bool = False,
) -> List[DummyTxPayload]:
    """Returns a list of DummyTxPayload objects."""
    return [
        DummyTxPayload(
            sender=agent,
            value=(value or agent) if not is_value_none else value,
            vote=vote if not is_vote_none else None,
        )
        for agent in sorted(participants)
    ]


class DummyRound(AbstractRound):
    """Dummy round."""

    payload_class = DummyTxPayload
    payload_attribute = "value"
    synchronized_data_class = BaseSynchronizedData

    def end_block(self) -> Optional[Tuple[BaseSynchronizedData, Enum]]:
        """end_block method."""


class DummyCollectionRound(CollectionRound, DummyRound):
    """Dummy Class for CollectionRound"""


class DummyCollectDifferentUntilAllRound(CollectDifferentUntilAllRound, DummyRound):
    """Dummy Class for CollectDifferentUntilAllRound"""


class DummyCollectSameUntilAllRound(CollectSameUntilAllRound, DummyRound):
    """Dummy Class for CollectSameUntilThresholdRound"""


class DummyCollectDifferentUntilThresholdRound(
    CollectDifferentUntilThresholdRound, DummyRound
):
    """Dummy Class for CollectDifferentUntilThresholdRound"""


class DummyCollectSameUntilThresholdRound(CollectSameUntilThresholdRound, DummyRound):
    """Dummy Class for CollectSameUntilThresholdRound"""


class DummyOnlyKeeperSendsRound(OnlyKeeperSendsRound, DummyRound):
    """Dummy Class for OnlyKeeperSendsRound"""

    fail_event = "FAIL_EVENT"


class DummyVotingRound(VotingRound, DummyRound):
    """Dummy Class for VotingRound"""


class DummyCollectNonEmptyUntilThresholdRound(
    CollectNonEmptyUntilThresholdRound, DummyRound
):
    """Dummy Class for `CollectNonEmptyUntilThresholdRound`"""


class BaseRoundTestClass:  # pylint: disable=too-few-public-methods
    """Base test class."""

    synchronized_data: BaseSynchronizedData
    participants: FrozenSet[str]

    _synchronized_data_class: Type[BaseSynchronizedData]
    _event_class: Any

    def setup(
        self,
    ) -> None:
        """Setup test class."""

        self.participants = get_participants()
        self.synchronized_data = self._synchronized_data_class(
            db=AbciAppDB(
                setup_data=dict(
                    participants=[tuple(self.participants)],
                    all_participants=[tuple(self.participants)],
                    consensus_threshold=[3],
                    safe_contract_address=["test_address"],
                ),
            )
        )

    def _test_no_majority_event(self, round_obj: AbstractRound) -> None:
        """Test the NO_MAJORITY event."""
        with mock.patch.object(round_obj, "is_majority_possible", return_value=False):
            result = round_obj.end_block()
            assert result is not None
            _, event = result
            assert event == self._event_class.NO_MAJORITY

    @staticmethod
    def _complete_run(
        test_runner: Generator, iter_count: int = MAX_PARTICIPANTS
    ) -> None:
        """
        This method represents logic to execute test logic defined in _test_round method.

        _test_round should follow these steps

        1. process first payload
        2. yield test_round
        3. test collection, end_block and thresholds
        4. process rest of the payloads
        5. yield test_round
        6. yield synchronized_data, event ( returned from end_block )
        7. test synchronized_data and event

        :param test_runner: test runner
        :param iter_count: iter_count
        """

        for _ in range(iter_count):
            next(test_runner)


class BaseCollectDifferentUntilAllRoundTest(  # pylint: disable=too-few-public-methods
    BaseRoundTestClass
):
    """Tests for rounds derived from CollectDifferentUntilAllRound."""

    def _test_round(
        self,
        test_round: CollectDifferentUntilAllRound,
        round_payloads: List[BaseTxPayload],
        synchronized_data_update_fn: Callable,
        synchronized_data_attr_checks: List[Callable],
        exit_event: Any,
    ) -> Generator:
        """Test round."""

        first_payload = round_payloads.pop(0)
        test_round.process_payload(first_payload)

        yield test_round
        assert test_round.collection[first_payload.sender] == first_payload
        assert not test_round.collection_threshold_reached
        assert test_round.end_block() is None

        for payload in round_payloads:
            test_round.process_payload(payload)
        yield test_round
        assert test_round.collection_threshold_reached

        actual_next_synchronized_data = synchronized_data_update_fn(
            deepcopy(self.synchronized_data), test_round
        )

        res = test_round.end_block()
        yield res
        if exit_event is None:
            assert res is exit_event
        else:
            assert res is not None
            synchronized_data, event = res
            for behaviour_attr_getter in synchronized_data_attr_checks:
                assert behaviour_attr_getter(
                    synchronized_data
                ) == behaviour_attr_getter(actual_next_synchronized_data)
            assert event == exit_event
        yield


class BaseCollectSameUntilAllRoundTest(
    BaseRoundTestClass
):  # pylint: disable=too-few-public-methods
    """Tests for rounds derived from CollectSameUntilAllRound."""

    def _test_round(  # pylint: disable=too-many-arguments,too-many-locals
        self,
        test_round: CollectSameUntilAllRound,
        round_payloads: Mapping[str, BaseTxPayload],
        synchronized_data_update_fn: Callable,
        synchronized_data_attr_checks: List[Callable],
        most_voted_payload: Any,
        exit_event: Any,
        finished: bool,
    ) -> Generator:
        """Test rounds derived from CollectionRound."""

        (_, first_payload), *payloads = round_payloads.items()

        test_round.process_payload(first_payload)
        yield test_round
        assert test_round.collection[first_payload.sender] == first_payload
        assert not test_round.collection_threshold_reached
        assert test_round.end_block() is None

        with pytest.raises(
            ABCIAppInternalError,
            match="internal error: 1 votes are not enough for `CollectSameUntilAllRound`. "
            f"Expected: `n_votes = max_participants = {MAX_PARTICIPANTS}`",
        ):
            _ = test_round.common_payload

        for _, payload in payloads:
            test_round.process_payload(payload)
        yield test_round
        if finished:
            assert test_round.collection_threshold_reached
        assert test_round.common_payload == most_voted_payload

        actual_next_synchronized_data = synchronized_data_update_fn(
            deepcopy(self.synchronized_data), test_round
        )
        res = test_round.end_block()
        yield res
        assert res is not None

        synchronized_data, event = res

        for behaviour_attr_getter in synchronized_data_attr_checks:
            assert behaviour_attr_getter(synchronized_data) == behaviour_attr_getter(
                actual_next_synchronized_data
            ), f"Mismatch in synchronized_data. Actual:\n{behaviour_attr_getter(synchronized_data)}\nExpected:\n{behaviour_attr_getter(actual_next_synchronized_data)}"
        assert event == exit_event
        yield


class BaseCollectSameUntilThresholdRoundTest(  # pylint: disable=too-few-public-methods
    BaseRoundTestClass
):
    """Tests for rounds derived from CollectSameUntilThresholdRound."""

    def _test_round(  # pylint: disable=too-many-arguments,too-many-locals
        self,
        test_round: CollectSameUntilThresholdRound,
        round_payloads: Mapping[str, BaseTxPayload],
        synchronized_data_update_fn: Callable,
        synchronized_data_attr_checks: List[Callable],
        most_voted_payload: Any,
        exit_event: Any,
    ) -> Generator:
        """Test rounds derived from CollectionRound."""

        (_, first_payload), *payloads = round_payloads.items()

        test_round.process_payload(first_payload)
        yield test_round
        assert test_round.collection[first_payload.sender] == first_payload
        assert not test_round.threshold_reached
        assert test_round.end_block() is None

        self._test_no_majority_event(test_round)
        with pytest.raises(ABCIAppInternalError, match="not enough votes"):
            _ = test_round.most_voted_payload

        for _, payload in payloads:
            test_round.process_payload(payload)
        yield test_round
        assert test_round.threshold_reached
        assert test_round.most_voted_payload == most_voted_payload

        actual_next_synchronized_data = synchronized_data_update_fn(
            deepcopy(self.synchronized_data), test_round
        )
        res = test_round.end_block()
        yield res
        assert res is not None

        synchronized_data, event = res

        for behaviour_attr_getter in synchronized_data_attr_checks:
            assert behaviour_attr_getter(synchronized_data) == behaviour_attr_getter(
                actual_next_synchronized_data
            ), f"Mismatch in synchronized_data. Actual:\n{behaviour_attr_getter(synchronized_data)}\nExpected:\n{behaviour_attr_getter(actual_next_synchronized_data)}"
        assert event == exit_event
        yield


class BaseOnlyKeeperSendsRoundTest(  # pylint: disable=too-few-public-methods
    BaseRoundTestClass
):
    """Tests for rounds derived from OnlyKeeperSendsRound."""

    def _test_round(
        self,
        test_round: OnlyKeeperSendsRound,
        keeper_payloads: BaseTxPayload,
        synchronized_data_update_fn: Callable,
        synchronized_data_attr_checks: List[Callable],
        exit_event: Any,
    ) -> Generator:
        """Test for rounds derived from OnlyKeeperSendsRound."""

        assert test_round.end_block() is None
        assert test_round.keeper_payload is None

        test_round.process_payload(keeper_payloads)
        yield test_round
        assert test_round.keeper_payload is not None

        yield test_round
        actual_next_synchronized_data = synchronized_data_update_fn(
            deepcopy(self.synchronized_data), test_round
        )
        res = test_round.end_block()
        yield res
        assert res is not None

        synchronized_data, event = res

        for behaviour_attr_getter in synchronized_data_attr_checks:
            assert behaviour_attr_getter(synchronized_data) == behaviour_attr_getter(
                actual_next_synchronized_data
            ), f"Mismatch in synchronized_data. Actual:\n{behaviour_attr_getter(synchronized_data)}\nExpected:\n{behaviour_attr_getter(actual_next_synchronized_data)}"
        assert event == exit_event
        yield


class BaseVotingRoundTest(BaseRoundTestClass):  # pylint: disable=too-few-public-methods
    """Tests for rounds derived from VotingRound."""

    def _test_round(  # pylint: disable=too-many-arguments,too-many-locals
        self,
        test_round: VotingRound,
        round_payloads: Mapping[str, BaseTxPayload],
        synchronized_data_update_fn: Callable,
        synchronized_data_attr_checks: List[Callable],
        exit_event: Any,
        threshold_check: Callable,
    ) -> Generator:
        """Test for rounds derived from VotingRound."""

        (_, first_payload), *payloads = round_payloads.items()

        test_round.process_payload(first_payload)
        yield test_round
        assert not threshold_check(test_round)  # negative_vote_threshold_reached
        assert test_round.end_block() is None
        self._test_no_majority_event(test_round)

        for _, payload in payloads:
            test_round.process_payload(payload)
        yield test_round
        assert threshold_check(test_round)

        actual_next_synchronized_data = synchronized_data_update_fn(
            deepcopy(self.synchronized_data), test_round
        )
        res = test_round.end_block()
        yield res
        assert res is not None

        synchronized_data, event = res

        for behaviour_attr_getter in synchronized_data_attr_checks:
            assert behaviour_attr_getter(synchronized_data) == behaviour_attr_getter(
                actual_next_synchronized_data
            ), f"Mismatch in synchronized_data. Actual:\n{behaviour_attr_getter(synchronized_data)}\nExpected:\n{behaviour_attr_getter(actual_next_synchronized_data)}"
        assert event == exit_event
        yield

    def _test_voting_round_positive(
        self,
        test_round: VotingRound,
        round_payloads: Mapping[str, BaseTxPayload],
        synchronized_data_update_fn: Callable,
        synchronized_data_attr_checks: List[Callable],
        exit_event: Any,
    ) -> Generator:
        """Test for rounds derived from VotingRound."""

        return self._test_round(
            test_round,
            round_payloads,
            synchronized_data_update_fn,
            synchronized_data_attr_checks,
            exit_event,
            threshold_check=lambda x: x.positive_vote_threshold_reached,
        )

    def _test_voting_round_negative(
        self,
        test_round: VotingRound,
        round_payloads: Mapping[str, BaseTxPayload],
        synchronized_data_update_fn: Callable,
        synchronized_data_attr_checks: List[Callable],
        exit_event: Any,
    ) -> Generator:
        """Test for rounds derived from VotingRound."""

        return self._test_round(
            test_round,
            round_payloads,
            synchronized_data_update_fn,
            synchronized_data_attr_checks,
            exit_event,
            threshold_check=lambda x: x.negative_vote_threshold_reached,
        )

    def _test_voting_round_none(
        self,
        test_round: VotingRound,
        round_payloads: Mapping[str, BaseTxPayload],
        synchronized_data_update_fn: Callable,
        synchronized_data_attr_checks: List[Callable],
        exit_event: Any,
    ) -> Generator:
        """Test for rounds derived from VotingRound."""

        return self._test_round(
            test_round,
            round_payloads,
            synchronized_data_update_fn,
            synchronized_data_attr_checks,
            exit_event,
            threshold_check=lambda x: x.none_vote_threshold_reached,
        )


class BaseCollectDifferentUntilThresholdRoundTest(  # pylint: disable=too-few-public-methods
    BaseRoundTestClass
):
    """Tests for rounds derived from CollectDifferentUntilThresholdRound."""

    def _test_round(
        self,
        test_round: CollectDifferentUntilThresholdRound,
        round_payloads: Mapping[str, BaseTxPayload],
        synchronized_data_update_fn: Callable,
        synchronized_data_attr_checks: List[Callable],
        exit_event: Any,
    ) -> Generator:
        """Test for rounds derived from CollectDifferentUntilThresholdRound."""

        (_, first_payload), *payloads = round_payloads.items()

        test_round.process_payload(first_payload)
        yield test_round
        assert not test_round.collection_threshold_reached
        assert test_round.end_block() is None

        for _, payload in payloads:
            test_round.process_payload(payload)
        yield test_round
        assert test_round.collection_threshold_reached

        actual_next_synchronized_data = synchronized_data_update_fn(
            deepcopy(self.synchronized_data), test_round
        )

        res = test_round.end_block()
        yield res
        assert res is not None

        synchronized_data, event = res

        for behaviour_attr_getter in synchronized_data_attr_checks:
            assert behaviour_attr_getter(synchronized_data) == behaviour_attr_getter(
                actual_next_synchronized_data
            ), f"Mismatch in synchronized_data. Actual:\n{behaviour_attr_getter(synchronized_data)}\nExpected:\n{behaviour_attr_getter(actual_next_synchronized_data)}"
        assert event == exit_event
        yield


class BaseCollectNonEmptyUntilThresholdRound(  # pylint: disable=too-few-public-methods
    BaseCollectDifferentUntilThresholdRoundTest
):
    """Tests for rounds derived from `CollectNonEmptyUntilThresholdRound`."""


class _BaseRoundTestClass(BaseRoundTestClass):  # pylint: disable=too-few-public-methods
    """Base test class."""

    synchronized_data: BaseSynchronizedData
    participants: FrozenSet[str]
    tx_payloads: List[DummyTxPayload]

    _synchronized_data_class = DummySynchronizedData

    def setup(
        self,
    ) -> None:
        """Setup test class."""

        super().setup()
        self.tx_payloads = get_dummy_tx_payloads(self.participants)

    @staticmethod
    def _test_payload_with_wrong_round_count(
        test_round: AbstractRound,
        value: Optional[str] = None,
        vote: Optional[bool] = None,
    ) -> None:
        """Test errors raised by payloads with wrong round count."""
        payload_with_wrong_round_count = DummyTxPayload("sender", value, vote)
        object.__setattr__(payload_with_wrong_round_count, "round_count", 0)
        with pytest.raises(
            TransactionNotValidError,
            match=re.escape("Expected round count -1 and got 0."),
        ):
            test_round.check_payload(payload=payload_with_wrong_round_count)

        with pytest.raises(
            ABCIAppInternalError,
            match=re.escape("Expected round count -1 and got 0."),
        ):
            test_round.process_payload(payload=payload_with_wrong_round_count)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test generators."""

from typing import List

from autonomy.fsm.scaffold.base import AbstractFileGenerator
from autonomy.fsm.scaffold.generators.components import SimpleFileGenerator
from autonomy.fsm.scaffold.templates import COPYRIGHT_HEADER
from autonomy.fsm.scaffold.templates.tests import (
    TEST_BEHAVIOURS,
    TEST_DIALOGUES,
    TEST_HANDLERS,
    TEST_MODELS,
    TEST_PAYLOADS,
    TEST_ROUNDS,
)


class RoundTestsFileGenerator(AbstractFileGenerator, TEST_ROUNDS):
    """RoundTestsFileGenerator"""

    def get_file_content(self) -> str:
        """Scaffold the 'test_rounds.py' file."""

        file_content = [
            COPYRIGHT_HEADER,
            self.HEADER.format(**self.template_kwargs),
            self._get_rounds_section(),
        ]

        return "\n".join(file_content)

    def _get_rounds_section(self) -> str:
        """Get rounds section"""

        rounds: List[str] = [self.BASE_ROUND_TEST_CLS.format(**self.template_kwargs)]

        for round_name in self.rounds:
            round_class_str = self.TEST_ROUND_CLS.format(
                FSMName=self.fsm_name,
                RoundCls=round_name,
            )
            rounds.append(round_class_str)

        return "\n".join(rounds)


class BehaviourTestsFileGenerator(AbstractFileGenerator, TEST_BEHAVIOURS):
    """File generator for 'test_behaviours.py' modules."""

    def get_file_content(self) -> str:
        """Scaffold the 'test_behaviours.py' file."""

        behaviour_file_content = [
            COPYRIGHT_HEADER,
            self.HEADER.format(**self.template_kwargs),
            self._get_behaviour_section(),
        ]

        return "\n".join(behaviour_file_content)

    def _get_behaviour_section(self) -> str:
        """Get behaviour section"""

        behaviours = [self.BASE_BEHAVIOUR_TEST_CLS.format(**self.template_kwargs)]

        for behaviour_name in self.behaviours:
            round_class_str = self.TEST_BEHAVIOUR_CLS.format(
                FSMName=self.fsm_name,
                BehaviourCls=behaviour_name,
            )
            behaviours.append(round_class_str)

        return "\n".join(behaviours)


class PayloadTestsFileGenerator(AbstractFileGenerator, TEST_PAYLOADS):
    """File generator for 'test_payloads.py' modules."""

    def get_file_content(self) -> str:
        """Scaffold the 'test_payloads.py' file."""

        behaviour_file_content = [
            COPYRIGHT_HEADER,
            self.HEADER.format(**self.template_kwargs),
            self.TEST_PAYLOAD_CLS.format(**self.template_kwargs),
        ]

        return "\n".join(behaviour_file_content)


class ModelTestFileGenerator(SimpleFileGenerator, TEST_MODELS):
    """File generator for 'test_models.py'."""


class HandlersTestFileGenerator(SimpleFileGenerator, TEST_HANDLERS):
    """File generator for 'test_dialogues.py'."""


class DialoguesTestFileGenerator(SimpleFileGenerator, TEST_DIALOGUES):
    """File generator for 'test_dialogues.py'."""



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""
This module contains a dummy ABCI skill in which the agents register and then pick up the termination signal.

It is created only for testing purposes for the termination feature.
"""  # pragma: nocover

from aea.configurations.base import PublicId  # pragma: nocover


PUBLIC_ID = PublicId.from_str("valory/register_termination_abci:0.1.0")



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Scaffolding templates for FSM"""

from datetime import datetime


COPYRIGHT_HEADER = """\
    # -*- coding: utf-8 -*-
    # ------------------------------------------------------------------------------
    #
    #   Copyright {year} Valory AG
    #
    #   Licensed under the Apache License, Version 2.0 (the "License");
    #   you may not use this file except in compliance with the License.
    #   You may obtain a copy of the License at
    #
    #       http://www.apache.org/licenses/LICENSE-2.0
    #
    #   Unless required by applicable law or agreed to in writing, software
    #   distributed under the License is distributed on an "AS IS" BASIS,
    #   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    #   See the License for the specific language governing permissions and
    #   limitations under the License.
    #
    # ------------------------------------------------------------------------------
    """.format(
    year=datetime.now().year
)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test the abci_app_chain.py module of the skill."""

# pylint: skip-file

import logging
from typing import Dict, Set, Tuple, Type
from unittest.mock import MagicMock

import pytest
from _pytest.logging import LogCaptureFixture
from aea.exceptions import AEAEnforceError

from packages.valory.skills.abstract_round_abci.abci_app_chain import (
    AbciAppTransitionMapping,
    chain,
)
from packages.valory.skills.abstract_round_abci.base import (
    AbciApp,
    AbciAppDB,
    AbstractRound,
    AppState,
    BaseSynchronizedData,
    BaseTxPayload,
    DegenerateRound,
)


def make_round_class(name: str, bases: Tuple = (AbstractRound,)) -> Type[AbstractRound]:
    """Make a round class."""
    new_round_cls = type(
        name,
        bases,
        {
            "synchronized_data_class": MagicMock(),
            "payload_class": MagicMock(),
            "payload_attribute": MagicMock(),
        },
    )
    setattr(new_round_cls, "round_id", name)  # noqa: B010
    assert issubclass(new_round_cls, AbstractRound)  # nosec
    return new_round_cls


class TestAbciAppChaining:
    """Test chaning of AbciApps."""

    def setup(self) -> None:
        """Setup test."""
        self.round_1a = make_round_class("round_1a")
        self.round_1b = make_round_class("round_1b")
        self.round_1b_dupe = make_round_class("round_1b")  # duplicated round id
        self.round_1c = make_round_class("round_1c", (DegenerateRound,))

        self.round_2a = make_round_class("round_2a")
        self.round_2b = make_round_class("round_2b")
        self.round_2c = make_round_class("round_2c", (DegenerateRound,))
        self.round_2d = make_round_class("round_2d")

        self.round_3a = make_round_class("round_3a")
        self.round_3b = make_round_class("round_3b")
        self.round_3c = make_round_class("round_3c", (DegenerateRound,))

        self.key_1 = "1"
        self.key_2 = "2"
        self.key_3 = "3"

        self.event_1a = "event_1a"
        self.event_1b = "event_1b"
        self.event_1c = "event_1c"
        self.event_timeout1 = "timeout_1"

        self.event_2a = "event_2a"
        self.event_2b = "event_2b"
        self.event_2c = "event_2c"
        self.event_timeout2 = "timeout_2"

        self.event_3a = "event_3a"
        self.event_3b = "event_3b"
        self.event_3c = "event_3c"
        self.event_timeout3 = "timeout_3"

        self.timeout1 = 10.0
        self.timeout2 = 15.0
        self.timeout3 = 20.0

        self.cross_period_persisted_keys_1 = frozenset({"1", "2"})
        self.cross_period_persisted_keys_2 = frozenset({"2", "3"})

        class AbciApp1(AbciApp):
            initial_round_cls = self.round_1a
            transition_function = {
                self.round_1a: {
                    self.event_timeout1: self.round_1a,
                    self.event_1b: self.round_1b,
                },
                self.round_1b: {
                    self.event_1a: self.round_1a,
                    self.event_1c: self.round_1c,
                },
                self.round_1c: {},
            }
            final_states = {self.round_1c}
            event_to_timeout = {self.event_timeout1: self.timeout1}
            db_pre_conditions: Dict[AppState, Set[str]] = {self.round_1a: set()}
            db_post_conditions: Dict[AppState, Set[str]] = {self.round_1c: {self.key_1}}
            cross_period_persisted_keys = self.cross_period_persisted_keys_1

        self.app1_class = AbciApp1

        class AbciApp2(AbciApp):
            initial_round_cls = self.round_2a
            transition_function = {
                self.round_2a: {
                    self.event_timeout2: self.round_2a,
                    self.event_2b: self.round_2b,
                },
                self.round_2b: {
                    self.event_2a: self.round_2a,
                    self.event_2c: self.round_2c,
                },
                self.round_2c: {},
            }
            final_states = {self.round_2c}
            event_to_timeout = {self.event_timeout2: self.timeout2}
            db_pre_conditions: Dict[AppState, Set[str]] = {self.round_2a: {self.key_1}}
            db_post_conditions: Dict[AppState, Set[str]] = {self.round_2c: {self.key_2}}
            cross_period_persisted_keys = self.cross_period_persisted_keys_2

        self.app2_class = AbciApp2

        class AbciApp3(AbciApp):
            initial_round_cls = self.round_3a
            transition_function = {
                self.round_3a: {
                    self.event_timeout3: self.round_3a,
                    self.event_3b: self.round_3b,
                },
                self.round_3b: {
                    self.event_3a: self.round_3a,
                    self.event_3c: self.round_3c,
                    self.event_1a: self.round_3a,  # duplicated event
                },
                self.round_3c: {},
            }
            final_states = {self.round_3c}
            event_to_timeout = {self.event_timeout3: self.timeout3}
            db_pre_conditions: Dict[AppState, Set[str]] = {
                self.round_3a: {self.key_1, self.key_2}
            }
            db_post_conditions: Dict[AppState, Set[str]] = {self.round_3c: {self.key_3}}

        self.app3_class = AbciApp3

        class AbciApp3Dupe(AbciApp):
            initial_round_cls = self.round_3a
            transition_function = {
                self.round_3a: {
                    self.event_timeout3: self.round_3a,
                    self.event_3b: self.round_3b,
                },
                self.round_1b_dupe: {  # duplucated round id
                    self.event_3a: self.round_3a,
                    self.event_3c: self.round_3c,
                    self.event_1a: self.round_3a,  # duplicated event
                },
                self.round_3c: {},
            }
            final_states = {self.round_3c}
            event_to_timeout = {self.event_timeout3: self.timeout3}
            db_post_conditions: Dict[AppState, Set[str]] = {self.round_3c: set()}

        self.app3_class_dupe = AbciApp3Dupe

        class AbciApp2Faulty1(AbciApp):
            initial_round_cls = self.round_2a
            transition_function = {
                self.round_2a: {
                    self.event_timeout2: self.round_2a,
                    self.event_2b: self.round_2b,
                },
                self.round_2b: {
                    self.event_2a: self.round_2a,
                    self.event_2c: self.round_2c,
                },
                self.round_2c: {},
            }
            final_states = {self.round_2c}
            event_to_timeout = {self.event_timeout1: self.timeout2}
            db_pre_conditions: Dict[AppState, Set[str]] = {self.round_2a: {self.key_1}}
            db_post_conditions: Dict[AppState, Set[str]] = {self.round_2c: {self.key_2}}

        self.app2_class_faulty1 = AbciApp2Faulty1

    def test_chain_two(self) -> None:
        """Test the AbciApp chain function."""

        abci_app_transition_mapping: AbciAppTransitionMapping = {
            self.round_1c: self.round_2a,
            self.round_2c: self.round_1a,
        }

        ComposedAbciApp = chain(
            (self.app1_class, self.app2_class), abci_app_transition_mapping
        )

        assert ComposedAbciApp.initial_round_cls == self.round_1a
        assert ComposedAbciApp.transition_function == {
            self.round_1a: {
                self.event_timeout1: self.round_1a,
                self.event_1b: self.round_1b,
            },
            self.round_1b: {
                self.event_1a: self.round_1a,
                self.event_1c: self.round_2a,
            },
            self.round_2a: {
                self.event_timeout2: self.round_2a,
                self.event_2b: self.round_2b,
            },
            self.round_2b: {
                self.event_2a: self.round_2a,
                self.event_2c: self.round_1a,
            },
        }
        assert ComposedAbciApp.final_states == set()
        assert ComposedAbciApp.event_to_timeout == {
            self.event_timeout1: self.timeout1,
            self.event_timeout2: self.timeout2,
        }
        assert (
            ComposedAbciApp.cross_period_persisted_keys
            == self.cross_period_persisted_keys_1.union(
                self.cross_period_persisted_keys_2
            )
        )

    def test_chain_three(self) -> None:
        """Test the AbciApp chain function."""

        abci_app_transition_mapping: AbciAppTransitionMapping = {
            self.round_1c: self.round_2a,
            self.round_2c: self.round_3a,
        }

        ComposedAbciApp = chain(
            (self.app1_class, self.app2_class, self.app3_class),
            abci_app_transition_mapping,
        )

        assert ComposedAbciApp.initial_round_cls == self.round_1a
        assert ComposedAbciApp.transition_function == {
            self.round_1a: {
                self.event_timeout1: self.round_1a,
                self.event_1b: self.round_1b,
            },
            self.round_1b: {
                self.event_1a: self.round_1a,
                self.event_1c: self.round_2a,
            },
            self.round_2a: {
                self.event_timeout2: self.round_2a,
                self.event_2b: self.round_2b,
            },
            self.round_2b: {
                self.event_2a: self.round_2a,
                self.event_2c: self.round_3a,
            },
            self.round_3a: {
                self.event_timeout3: self.round_3a,
                self.event_3b: self.round_3b,
            },
            self.round_3b: {
                self.event_3a: self.round_3a,
                self.event_3c: self.round_3c,
                self.event_1a: self.round_3a,
            },
            self.round_3c: {},
        }
        assert ComposedAbciApp.final_states == {self.round_3c}
        assert ComposedAbciApp.event_to_timeout == {
            self.event_timeout1: self.timeout1,
            self.event_timeout2: self.timeout2,
            self.event_timeout3: self.timeout3,
        }

    def test_chain_two_negative_timeouts(self) -> None:
        """Test the AbciApp chain function."""

        abci_app_transition_mapping: AbciAppTransitionMapping = {
            self.round_1c: self.round_2a,
            self.round_2c: self.round_1a,
        }

        with pytest.raises(
            ValueError, match="but it is already defined in a prior app with timeout"
        ):
            _ = chain(
                (self.app1_class, self.app2_class_faulty1), abci_app_transition_mapping
            )

    def test_chain_two_negative_mapping_initial_states(self) -> None:
        """Test the AbciApp chain function."""

        abci_app_transition_mapping: AbciAppTransitionMapping = {
            self.round_1c: self.round_2b,
            self.round_2c: self.round_1a,
        }

        with pytest.raises(ValueError, match="Found non-initial state"):
            _ = chain((self.app1_class, self.app2_class), abci_app_transition_mapping)

    def test_chain_two_negative_mapping_final_states(self) -> None:
        """Test the AbciApp chain function."""

        abci_app_transition_mapping: AbciAppTransitionMapping = {
            self.round_1c: self.round_2a,
            self.round_2b: self.round_1a,
        }

        with pytest.raises(ValueError, match="Found non-final state"):
            _ = chain((self.app1_class, self.app2_class), abci_app_transition_mapping)

    def test_chain_two_dupe(self) -> None:
        """Test the AbciApp chain function."""

        abci_app_transition_mapping: AbciAppTransitionMapping = {
            self.round_1c: self.round_2a,
            self.round_2c: self.round_1a,
        }
        with pytest.raises(
            AEAEnforceError,
            match=r"round ids in common between abci apps are not allowed.*",
        ):
            chain((self.app1_class, self.app3_class_dupe), abci_app_transition_mapping)

    def test_chain_with_abstract_abci_app_fails(self) -> None:
        """Test chaining with an abstract AbciApp fails."""
        self.app2_class._is_abstract = False
        self.app3_class._is_abstract = False
        with pytest.raises(
            AEAEnforceError,
            match=r"found non-abstract AbciApp during chaining: \['AbciApp2', 'AbciApp3'\]",
        ):
            abci_app_transition_mapping: AbciAppTransitionMapping = {
                self.round_1c: self.round_2a,
                self.round_2c: self.round_3a,
            }
            chain(
                (self.app1_class, self.app2_class, self.app3_class),
                abci_app_transition_mapping,
            )

    def test_synchronized_data_type(self, caplog: LogCaptureFixture) -> None:
        """Test synchronized data type"""

        abci_app_transition_mapping: AbciAppTransitionMapping = {
            self.round_1c: self.round_2a,
            self.round_2c: self.round_1a,
        }

        sentinel_app1 = object()
        sentinel_app2 = object()

        def make_sync_data(sentinel: object) -> Type:
            class SynchronizedData(BaseSynchronizedData):
                @property
                def dummy_attr(self) -> object:
                    return sentinel

            return SynchronizedData

        def make_concrete(round_cls: Type[AbstractRound]) -> Type[AbstractRound]:
            class ConcreteRound(round_cls):  # type: ignore
                def check_payload(self, payload: BaseTxPayload) -> None:
                    pass

                def process_payload(self, payload: BaseTxPayload) -> None:
                    pass

                def end_block(self) -> None:
                    pass

                payload_class = None

            return ConcreteRound

        sync_data_cls_app1 = make_sync_data(sentinel_app1)
        sync_data_cls_app2 = make_sync_data(sentinel_app2)

        # don't need to mock all of this since setup creates these classes
        for abci_app_cls, sync_data_cls in (
            (self.app1_class, sync_data_cls_app1),
            (self.app2_class, sync_data_cls_app2),
        ):
            synchronized_data = sync_data_cls(db=AbciAppDB(setup_data={}))
            abci_app = abci_app_cls(synchronized_data, logging.getLogger())
            for r in abci_app_cls.get_all_rounds():
                r.synchronized_data_class = sync_data_cls

        abci_app_cls = chain(
            (self.app1_class, self.app2_class), abci_app_transition_mapping
        )
        synchronized_data = sync_data_cls_app2(db=AbciAppDB(setup_data={}))
        abci_app = abci_app_cls(synchronized_data, logging.getLogger())

        assert abci_app.initial_round_cls == self.round_1a
        assert isinstance(abci_app.synchronized_data, sync_data_cls_app1)
        assert abci_app.synchronized_data.dummy_attr == sentinel_app1

        app2_classes = self.app2_class.get_all_rounds()
        for round_ in sorted(abci_app.get_all_rounds(), key=str):
            abci_app._round_results.append(abci_app.synchronized_data)
            abci_app.schedule_round(make_concrete(round_))
            expected_cls = (sync_data_cls_app1, sync_data_cls_app2)[
                round_ in app2_classes
            ]
            assert isinstance(abci_app.synchronized_data, expected_cls)
            expected_sentinel = (sentinel_app1, sentinel_app2)[round_ in app2_classes]
            assert abci_app.synchronized_data.dummy_attr == expected_sentinel

    def test_precondition_for_next_app_missing_raises(
        self, caplog: LogCaptureFixture
    ) -> None:
        """Test that when precondition for next AbciApp is missing an error is raised"""

        class AbciApp1(AbciApp):
            initial_round_cls = self.round_1a
            transition_function = {
                self.round_1a: {
                    self.event_timeout1: self.round_1a,
                    self.event_1b: self.round_1b,
                },
                self.round_1b: {
                    self.event_1a: self.round_1a,
                    self.event_1c: self.round_1c,
                },
                self.round_1c: {},
            }
            final_states = {self.round_1c}
            event_to_timeout = {self.event_timeout1: self.timeout1}
            db_pre_conditions: Dict[AppState, Set[str]] = {self.round_1a: set()}
            db_post_conditions: Dict[AppState, Set[str]] = {self.round_1c: set()}
            cross_period_persisted_keys = self.cross_period_persisted_keys_1

        abci_app_transition_mapping: AbciAppTransitionMapping = {
            self.round_1c: self.round_2a,
            self.round_2c: self.round_1a,
        }

        expected = "Pre conditions '.*' of app '.*' not a post condition of app '.*' or any preceding app in path .*."
        with pytest.raises(ValueError, match=expected):
            chain(
                (
                    AbciApp1,
                    self.app2_class,
                ),
                abci_app_transition_mapping,
            )

    def test_precondition_app_missing_raises(self, caplog: LogCaptureFixture) -> None:
        """Test that missing precondition specification for next AbciApp is missing an error is raised"""

        class AbciApp2(AbciApp):
            initial_round_cls = self.round_2a
            transition_function = {
                self.round_2a: {
                    self.event_timeout2: self.round_2a,
                    self.event_2b: self.round_2b,
                },
                self.round_2b: {
                    self.event_2a: self.round_2a,
                    self.event_2c: self.round_2c,
                },
                self.round_2c: {},
            }
            final_states = {self.round_2c}
            event_to_timeout = {self.event_timeout2: self.timeout2}
            db_pre_conditions: Dict[AppState, Set[str]] = {}
            db_post_conditions: Dict[AppState, Set[str]] = {self.round_2c: {self.key_2}}
            cross_period_persisted_keys = self.cross_period_persisted_keys_2

        abci_app_transition_mapping: AbciAppTransitionMapping = {
            self.round_1c: self.round_2a,
            self.round_2c: self.round_1a,
        }

        expected = "No pre-conditions have been set for .*! You need to explicitly specify them as empty if there are no pre-conditions for this FSM."
        with pytest.raises(ValueError, match=expected):
            chain((self.app1_class, AbciApp2), abci_app_transition_mapping)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""A connection responsible for uploading and downloading files from IPFS."""
import asyncio
import os
import tempfile
from asyncio import Task
from concurrent.futures import Executor
from pathlib import Path
from shutil import rmtree
from typing import Any, Callable, Dict, Optional, cast

import requests
from aea.configurations.base import PublicId
from aea.connections.base import Connection, ConnectionStates
from aea.mail.base import Envelope
from aea.protocols.base import Address, Message
from aea.protocols.dialogue.base import Dialogue as BaseDialogue
from aea_cli_ipfs.exceptions import DownloadError
from aea_cli_ipfs.ipfs_utils import IPFSTool
from ipfshttpclient.exceptions import ErrorResponse

from packages.valory.protocols.ipfs import IpfsMessage
from packages.valory.protocols.ipfs.dialogues import IpfsDialogue
from packages.valory.protocols.ipfs.dialogues import IpfsDialogues as BaseIpfsDialogues


PUBLIC_ID = PublicId.from_str("valory/ipfs:0.1.0")


class IpfsDialogues(BaseIpfsDialogues):
    """A class to keep track of IPFS dialogues."""

    def __init__(self, **kwargs: Any) -> None:
        """
        Initialize dialogues.

        :param kwargs: keyword arguments
        """

        def role_from_first_message(  # pylint: disable=unused-argument
            message: Message, receiver_address: Address
        ) -> BaseDialogue.Role:
            """Infer the role of the agent from an incoming/outgoing first message

            :param message: an incoming/outgoing first message
            :param receiver_address: the address of the receiving agent
            :return: The role of the agent
            """
            return IpfsDialogue.Role.CONNECTION

        BaseIpfsDialogues.__init__(
            self,
            self_address=str(kwargs.pop("connection_id")),
            role_from_first_message=role_from_first_message,
            **kwargs,
        )


class IpfsConnection(Connection):
    """An async connection for sending and receiving files to IPFS."""

    connection_id = PUBLIC_ID

    def __init__(self, **kwargs: Any) -> None:
        """
        Initialize the connection.

        :param kwargs: keyword arguments passed to component base.
        """
        super().__init__(**kwargs)  # pragma: no cover
        ipfs_domain = self.configuration.config.get("ipfs_domain")
        self.ipfs_tool: IPFSTool = IPFSTool(ipfs_domain)
        self.task_to_request: Dict[asyncio.Future, Envelope] = {}
        self.loop_executor: Optional[Executor] = None
        self.dialogues = IpfsDialogues(connection_id=PUBLIC_ID)
        self._response_envelopes: Optional[asyncio.Queue] = None

    @property
    def response_envelopes(self) -> asyncio.Queue:
        """Returns the response envelopes queue."""
        if self._response_envelopes is None:
            raise ValueError(
                "`IPFSConnection.response_envelopes` is not yet initialized. Is the connection setup?"
            )
        return self._response_envelopes

    async def connect(self) -> None:
        """Set up the connection."""
        self.ipfs_tool.check_ipfs_node_running()
        self._response_envelopes = asyncio.Queue()
        self.state = ConnectionStates.connected

    async def disconnect(self) -> None:
        """Tear down the connection."""
        if self.is_disconnected:  # pragma: nocover
            return

        self.state = ConnectionStates.disconnecting

        for task in self.task_to_request.keys():
            if not task.cancelled():  # pragma: nocover
                task.cancel()
        self._response_envelopes = None

        self.state = ConnectionStates.disconnected

    async def send(self, envelope: Envelope) -> None:
        """Send an envelope."""
        task = self._handle_envelope(envelope)
        task.add_done_callback(self._handle_done_task)
        self.task_to_request[task] = envelope

    async def receive(self, *args: Any, **kwargs: Any) -> Optional[Envelope]:
        """Receive an envelope."""
        return await self.response_envelopes.get()

    def run_async(
        self, func: Callable, *args: Any, timeout: Optional[float] = None
    ) -> Task:
        """Run a function asynchronously by using threads."""
        ipfs_operation = self.loop.run_in_executor(
            self.loop_executor,
            func,
            *args,
        )
        timely_operation = asyncio.wait_for(ipfs_operation, timeout=timeout)
        task = self.loop.create_task(timely_operation)
        return task

    def _handle_envelope(self, envelope: Envelope) -> Task:
        """Handle incoming envelopes by dispatching background tasks."""
        message = cast(IpfsMessage, envelope.message)
        performative = message.performative
        handler = getattr(self, f"_handle_{performative.value}", None)
        if handler is None:
            err = f"Performative `{performative.value}` is not supported."
            self.logger.error(err)
            task = self.run_async(self._handle_error, err)
            return task
        dialogue = self.dialogues.update(message)
        task = self.run_async(handler, message, dialogue)
        return task

    def _handle_store_files(
        self, message: IpfsMessage, dialogue: BaseDialogue
    ) -> IpfsMessage:
        """
        Handle a STORE_FILES performative.

        Uploads the provided files to ipfs.

        :param message: The ipfs request.
        :returns: the hash of the uploaded files.
        """
        files = message.files
        if len(files) == 0:
            err = "No files were present."
            self.logger.error(err)
            return self._handle_error(err, dialogue)
        if len(files) == 1:
            # a single file needs to be stored,
            # we don't need to create a dir
            path, data = files.popitem()
            self.__create_file(path, data)
        else:
            # multiple files are present, which means that it's a directory
            # we begin by checking that they belong to the same directory
            dirs = {os.path.dirname(path) for path in files.keys()}
            if len(dirs) > 1:
                err = f"Received files from different dirs {dirs}. "
                self.logger.error(err)
                self.logger.info(
                    "If you want to send multiple files as a single dir, "
                    "make sure the their path matches to one directory only."
                )
                return self._handle_error(err, dialogue)

            # "path" is the directory, it's the same for all the files
            path = dirs.pop()
            os.makedirs(path, exist_ok=True)
            for file_path, data in files.items():
                self.__create_file(file_path, data)
        try:
            # note that path will be a dir if multiple files
            # are being uploaded.
            _, hash_, _ = self.ipfs_tool.add(path)
            self.logger.debug(f"Successfully stored files with hash: {hash_}.")
        except (
            ValueError,
            requests.exceptions.ChunkedEncodingError,
        ) as e:  # pragma: no cover
            err = str(e)
            self.logger.error(err)
            return self._handle_error(err, dialogue)
        finally:
            self.__remove_filepath(path)
        response_message = cast(
            IpfsMessage,
            dialogue.reply(
                performative=IpfsMessage.Performative.IPFS_HASH,
                target_message=message,
                ipfs_hash=hash_,
            ),
        )
        return response_message

    def _handle_get_files(
        self, message: IpfsMessage, dialogue: BaseDialogue
    ) -> IpfsMessage:
        """
        Handle GET_FILES performative.

        Downloads and returns the files resulting from the ipfs hash.

        :param message: The ipfs request.
        :returns: the downloaded files.
        """
        response_body: Dict[str, str] = {}
        ipfs_hash = message.ipfs_hash
        with tempfile.TemporaryDirectory() as tmp_dir:
            try:
                self.ipfs_tool.download(ipfs_hash, tmp_dir)
            except (
                DownloadError,
                PermissionError,
                ErrorResponse,
            ) as e:
                err = str(e)
                self.logger.error(err)
                return self._handle_error(err, dialogue)

            files = os.listdir(tmp_dir)
            if len(files) > 1:
                self.logger.warning(
                    f"Multiple files or dirs found in {tmp_dir}. "
                    f"The first will be used. "
                )
            downloaded_file = files.pop()
            files_to_be_read = [downloaded_file]
            base_dir = Path(tmp_dir)
            if os.path.isdir(base_dir / downloaded_file):
                base_dir = base_dir / downloaded_file
                files_to_be_read = os.listdir(base_dir)

            for file_path in files_to_be_read:
                with open(base_dir / file_path, encoding="utf-8", mode="r") as file:
                    response_body[file_path] = file.read()

            response_message = cast(
                IpfsMessage,
                dialogue.reply(
                    performative=IpfsMessage.Performative.FILES,
                    files=response_body,
                    target_message=message,
                ),
            )
            return response_message

    def _handle_error(  # pylint: disable=no-self-use
        self, reason: str, dialogue: BaseDialogue
    ) -> IpfsMessage:
        """Handler for error messages."""
        message = cast(
            IpfsMessage,
            dialogue.reply(
                performative=IpfsMessage.Performative.ERROR,
                reason=reason,
            ),
        )
        return message

    def _handle_done_task(self, task: asyncio.Future) -> None:
        """
        Process a done receiving task.

        :param task: the done task.
        """
        request = self.task_to_request.pop(task)
        response_message: Optional[Message] = task.result()

        response_envelope = None
        if response_message is not None:
            response_envelope = Envelope(
                to=request.sender,
                sender=request.to,
                message=response_message,
                context=request.context,
            )

        # not handling `asyncio.QueueFull` exception, because the maxsize we defined for the Queue is infinite
        self.response_envelopes.put_nowait(response_envelope)

    @staticmethod
    def __create_file(path: str, data: str) -> None:
        """Creates a file in the specified path."""
        with open(path, "w", encoding="utf-8") as f:
            f.write(data)

    def __remove_filepath(self, filepath: str) -> None:
        """Remove a file or a folder. If filepath is not a file or a folder, an `IPFSInteractionError` is raised."""
        if os.path.isfile(filepath):
            os.remove(filepath)
        elif os.path.isdir(filepath):
            rmtree(filepath)
        else:  # pragma: no cover
            self.logger.error(f"`{filepath}` is not an existing filepath!")



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Metadata helpers."""

import json
from collections import OrderedDict
from pathlib import Path
from typing import Tuple, Union

from aea.configurations.data_types import PackageId
from aea.helpers.base import IPFSHash
from aea.helpers.cid import CID, to_v1
from aea.helpers.ipfs.base import IPFSHashOnly
from aea_cli_ipfs.ipfs_utils import IPFSTool


IPFS_URI_PREFIX = "ipfs://"
BASE16_HASH_PREFIX = "f01701220"
CONFIG_HASH_STRING_PREFIX = "0x"
UNIT_HASH_PREFIX = CONFIG_HASH_STRING_PREFIX + "{metadata_hash}"

NFTHashOrPath = Union[Path, IPFSHash]


def serialize_metadata(
    package_hash: str,
    package_id: PackageId,
    description: str,
    nft_image_hash: str,
) -> str:
    """Serialize metadata."""
    metadata = OrderedDict(
        {
            "name": package_id.to_uri_path,
            "description": description,
            "code_uri": f"{IPFS_URI_PREFIX}{package_hash}",
            "image": f"ipfs://{nft_image_hash}",
            "attributes": [{"trait_type": "version", "value": str(package_id.version)}],
        }
    )
    return json.dumps(obj=metadata, separators=(",", ":"))


def publish_metadata(
    package_id: PackageId,
    package_path: Path,
    nft: NFTHashOrPath,
    description: str,
) -> Tuple[str, str]:
    """Publish service metadata."""
    ipfs_tool = IPFSTool()
    if isinstance(nft, IPFSHash):
        nft_image_hash = str(nft)
    else:
        _, nft_image_hash, _ = ipfs_tool.add(
            str(nft.resolve()),
            wrap_with_directory=False,
        )
        nft_image_hash = to_v1(nft_image_hash)

    package_hash = IPFSHashOnly.get(file_path=str(package_path))
    metadata_string = serialize_metadata(
        package_hash=package_hash,
        package_id=package_id,
        description=description,
        nft_image_hash=nft_image_hash,
    )

    metadata_hash = ipfs_tool.client.add_str(metadata_string)
    metadata_hash = (
        CID.from_string(metadata_hash)
        .to_v1()
        .encode("base16")
        .decode()
        .replace(BASE16_HASH_PREFIX, "")
    )
    return UNIT_HASH_PREFIX.format(metadata_hash=metadata_hash), metadata_string



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Tests for abstract_round_abci/test_tools/integration.py"""

from typing import cast

import pytest

from packages.open_aea.protocols.signing import SigningMessage
from packages.open_aea.protocols.signing.custom_types import SignedMessage
from packages.valory.connections.ledger.connection import (
    PUBLIC_ID as LEDGER_CONNECTION_PUBLIC_ID,
)
from packages.valory.connections.ledger.tests.conftest import make_ledger_api_connection
from packages.valory.protocols.ledger_api import LedgerApiMessage
from packages.valory.protocols.ledger_api.dialogues import LedgerApiDialogue
from packages.valory.skills.abstract_round_abci.base import AbciAppDB
from packages.valory.skills.abstract_round_abci.behaviours import BaseBehaviour
from packages.valory.skills.abstract_round_abci.models import Requests
from packages.valory.skills.abstract_round_abci.test_tools.integration import (
    IntegrationBaseCase,
)
from packages.valory.skills.abstract_round_abci.tests.data.dummy_abci.behaviours import (
    DummyStartingBehaviour,
)
from packages.valory.skills.abstract_round_abci.tests.data.dummy_abci.rounds import (
    SynchronizedData,
)
from packages.valory.skills.abstract_round_abci.tests.test_tools.base import (
    FSMBehaviourTestToolSetup,
)


def simulate_ledger_get_balance_request(test_instance: IntegrationBaseCase) -> None:
    """Simulate ledger GET_BALANCE request"""

    ledger_api_dialogues = test_instance.skill.skill_context.ledger_api_dialogues
    ledger_api_msg, ledger_api_dialogue = ledger_api_dialogues.create(
        counterparty=str(LEDGER_CONNECTION_PUBLIC_ID),
        performative=LedgerApiMessage.Performative.GET_BALANCE,
        ledger_id="ethereum",
        address="0x" + "0" * 40,
    )
    ledger_api_dialogue = cast(LedgerApiDialogue, ledger_api_dialogue)
    current_behaviour = cast(BaseBehaviour, test_instance.behaviour.current_behaviour)
    request_nonce = current_behaviour._get_request_nonce_from_dialogue(  # pylint: disable=protected-access
        ledger_api_dialogue
    )
    cast(Requests, current_behaviour.context.requests).request_id_to_callback[
        request_nonce
    ] = current_behaviour.get_callback_request()
    current_behaviour.context.outbox.put_message(message=ledger_api_msg)


class TestIntegrationBaseCase(FSMBehaviourTestToolSetup):
    """TestIntegrationBaseCase"""

    test_cls = IntegrationBaseCase

    def test_instantiation(self) -> None:
        """Test instantiation"""

        self.set_path_to_skill()
        self.test_cls.make_ledger_api_connection_callable = make_ledger_api_connection
        test_instance = cast(IntegrationBaseCase, self.setup_test_cls())

        assert test_instance
        assert test_instance.get_message_from_outbox() is None
        assert test_instance.get_message_from_decision_maker_inbox() is None
        assert test_instance.process_n_messages(ncycles=0) is tuple()

        expected = "Invalid number of messages in outbox. Expected 1. Found 0."
        with pytest.raises(AssertionError, match=expected):
            assert test_instance.process_message_cycle()
        with pytest.raises(AssertionError, match=expected):
            assert test_instance.process_n_messages(ncycles=1)

    def test_process_messages_cycle(self) -> None:
        """Test process_message_cycle"""

        self.set_path_to_skill()
        self.test_cls.make_ledger_api_connection_callable = make_ledger_api_connection
        test_instance = cast(IntegrationBaseCase, self.setup_test_cls())

        simulate_ledger_get_balance_request(test_instance)
        message = test_instance.process_message_cycle(
            handler=None,
        )
        assert message is None

        simulate_ledger_get_balance_request(test_instance)
        # connection error - cannot dynamically mix in an autouse fixture
        message = test_instance.process_message_cycle(
            handler=test_instance.ledger_handler,
            expected_content={"performative": LedgerApiMessage.Performative.ERROR},
        )
        assert message

    def test_process_n_messages(self) -> None:
        """Test process_n_messages"""

        self.set_path_to_skill()
        self.test_cls.make_ledger_api_connection_callable = make_ledger_api_connection
        test_instance = cast(IntegrationBaseCase, self.setup_test_cls())

        behaviour_id = DummyStartingBehaviour.auto_behaviour_id()
        synchronized_data = SynchronizedData(
            AbciAppDB(setup_data=dict(participants=[tuple("abcd")]))
        )

        handlers = [test_instance.signing_handler]
        expected_content = [
            {"performative": SigningMessage.Performative.SIGNED_MESSAGE}
        ]
        expected_types = [{"signed_message": SignedMessage}]

        messages = test_instance.process_n_messages(
            ncycles=1,
            behaviour_id=behaviour_id,
            synchronized_data=synchronized_data,
            handlers=handlers,  # type: ignore
            expected_content=expected_content,  # type: ignore
            expected_types=expected_types,  # type: ignore
            fail_send_a2a=True,
        )
        assert len(messages) == 1



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains all the interaction operations of the behaviours with IPFS."""


import os
from typing import Any, Dict, Optional, Type

from packages.valory.skills.abstract_round_abci.io_.load import (
    CustomLoaderType,
    Loader,
    SupportedFiletype,
    SupportedObjectType,
)
from packages.valory.skills.abstract_round_abci.io_.store import (
    CustomStorerType,
    Storer,
)


class IPFSInteractionError(Exception):
    """A custom exception for IPFS interaction errors."""


class IPFSInteract:
    """Class for interacting with IPFS."""

    def __init__(self, loader_cls: Type = Loader, storer_cls: Type = Storer):
        """Initialize an `IPFSInteract` object."""
        # Set loader/storer class.
        self._loader_cls = loader_cls
        self._storer_cls = storer_cls

    def store(
        self,
        filepath: str,
        obj: SupportedObjectType,
        multiple: bool,
        filetype: Optional[SupportedFiletype] = None,
        custom_storer: Optional[CustomStorerType] = None,
        **kwargs: Any,
    ) -> Dict[str, str]:
        """Temporarily store a file locally, in order to send it to IPFS and retrieve a hash, and then delete it."""
        filepath = os.path.normpath(filepath)
        if multiple:
            # Add trailing slash in order to treat path as a folder.
            filepath = os.path.join(filepath, "")
        storer = self._storer_cls(filetype, custom_storer, filepath)

        try:
            name_to_obj = storer.store(obj, multiple, **kwargs)
            return name_to_obj
        except Exception as e:  # pylint: disable=broad-except
            raise IPFSInteractionError(str(e)) from e

    def load(  # pylint: disable=too-many-arguments
        self,
        serialized_objects: Dict[str, str],
        filetype: Optional[SupportedFiletype] = None,
        custom_loader: CustomLoaderType = None,
    ) -> SupportedObjectType:
        """Deserialize objects received via IPFS."""
        loader = self._loader_cls(filetype, custom_loader)
        try:
            deserialized_objects = loader.load(serialized_objects)
            return deserialized_objects
        except Exception as e:  # pylint: disable=broad-except
            raise IPFSInteractionError(str(e)) from e



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Service component base."""

import os
from pathlib import Path
from typing import Dict
from warnings import warn

from aea.configurations.base import (
    ConnectionConfig,
    ContractConfig,
    ProtocolConfig,
    SkillConfig,
)
from aea.helpers.env_vars import apply_env_variables
from aea.helpers.io import open_file
from aea.helpers.yaml_utils import yaml_load_all

from autonomy.configurations.base import Service


COMPONENT_CONFIGS: Dict = {
    component.package_type.value: component  # type: ignore
    for component in [
        ContractConfig,
        SkillConfig,
        ProtocolConfig,
        ConnectionConfig,
    ]
}


def load_service_config(
    service_path: Path, substitute_env_vars: bool = False
) -> Service:
    """Load service config from the path."""

    if substitute_env_vars:
        warn(
            "`substitute_env_vars` argument is deprecated and will be removed in v1.0.0, "
            "usage of environment varibales is default now.",
            DeprecationWarning,
            stacklevel=2,
        )

    # TODO: align with open-aea _load_service_config & handle data == None and other validation errors
    with open_file(
        service_path / Service.default_configuration_filename, "r", encoding="utf-8"
    ) as fp:
        data = yaml_load_all(fp)

    # Here we apply the environment variables to base service config only
    # We apply the environment variables to the overrides when processing
    # them to export as environment variables
    service_config, *overrides = data
    service_config = apply_env_variables(
        service_config, env_variables=os.environ.copy()
    )
    Service.validate_config_data(service_config)
    service_config["license_"] = service_config.pop("license")

    service = Service(**service_config)
    service.overrides = overrides

    return service



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Tests for valory/multicall2 contract."""
from pathlib import Path
from typing import Any, Dict, cast

from aea_test_autonomy.base_test_classes.contracts import BaseGanacheContractTest
from aea_test_autonomy.configurations import DEFAULT_AMOUNT as DEFAULT_ETH_BALANCE
from aea_test_autonomy.docker.base import skip_docker_tests

from packages.valory.contracts.multicall2.contract import Multicall2Contract


DEFAULT_GAS = 10000000


@skip_docker_tests
class TestTokenSettingsFactory(BaseGanacheContractTest):
    """Test deployment of Token Settings to Ganache."""

    contract_directory = Path(__file__).parent.parent
    contract: Multicall2Contract

    @classmethod
    def deployment_kwargs(cls) -> Dict[str, Any]:
        """Get deployment kwargs."""
        return dict(
            gas=DEFAULT_GAS,
        )

    def test_aggregate(self) -> None:
        """Test aggregate."""
        address, _pk = self.key_pairs()[1]
        expected_funds = DEFAULT_ETH_BALANCE
        multicall_instance = Multicall2Contract.get_instance(
            self.ledger_api, self.contract_address
        )
        call = self.contract.encode_function_call(
            self.ledger_api,
            multicall_instance,
            fn_name="getEthBalance",
            args=[address],
        )
        # make the same call 3 times
        calls = [call, call, call]
        contract_address = cast(str, self.contract_address)
        results = self.contract.aggregate_and_decode(
            self.ledger_api, contract_address, calls
        )

        assert isinstance(results, tuple)
        assert len(results) == 2

        block_number, responses = results
        assert isinstance(block_number, int)
        assert isinstance(responses, list)
        assert len(responses) == len(calls)
        for response in responses:
            # the response always is Tuple
            assert isinstance(response, tuple)
            assert isinstance(response[0], int)
            actual_funds = response[0]
            assert actual_funds == expected_funds



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Tests for valory/register_reset_recovery skill's behaviours."""

from pathlib import Path
from typing import Any, Dict, Optional, Type
from unittest import mock

from packages.valory.skills.abstract_round_abci.base import (
    AbciAppDB,
    BaseSynchronizedData,
)
from packages.valory.skills.abstract_round_abci.behaviour_utils import BaseBehaviour
from packages.valory.skills.abstract_round_abci.test_tools.base import (
    FSMBehaviourBaseCase,
)
from packages.valory.skills.register_reset_recovery_abci.behaviours import (
    RegisterResetRecoveryAbciAppConsensusBehaviour,
    ShareRoundCountBehaviour,
)
from packages.valory.skills.register_reset_recovery_abci.rounds import Event


class BaseTerminationTest(FSMBehaviourBaseCase):
    """Base test case."""

    path_to_skill = Path(__file__).parents[1]

    behaviour: RegisterResetRecoveryAbciAppConsensusBehaviour
    behaviour_class: Type[BaseBehaviour]
    synchronized_data: BaseSynchronizedData
    next_behaviour_class = ShareRoundCountBehaviour
    done_event = Event.DONE

    def fast_forward(self, data: Optional[Dict[str, Any]] = None) -> None:
        """Fast-forward on initialization"""

        data = data if data is not None else {}
        self.fast_forward_to_behaviour(
            self.behaviour,
            self.behaviour_class.auto_behaviour_id(),
            BaseSynchronizedData(AbciAppDB(setup_data=AbciAppDB.data_to_lists(data))),
        )
        assert (
            self.behaviour.current_behaviour is not None
            and self.behaviour.current_behaviour.auto_behaviour_id()
            == self.behaviour_class.auto_behaviour_id()
        )

    def complete(self) -> None:
        """Complete test"""
        self.behaviour.act_wrapper()
        self.mock_a2a_transaction()
        self._test_done_flag_set()
        self.end_round(done_event=self.done_event)
        assert (
            self.behaviour.current_behaviour is not None
            and self.behaviour.current_behaviour.auto_behaviour_id()
            == self.next_behaviour_class.auto_behaviour_id()
        )


class TestShareRoundCountBehaviour(BaseTerminationTest):
    """Tests the ShareRoundCountBehaviour."""

    behaviour_class = ShareRoundCountBehaviour

    def test_run(self) -> None:
        """Run the behaviour."""

        self.fast_forward()
        with mock.patch.object(self.behaviour.current_behaviour, "sleep"):
            self.behaviour.act_wrapper()
            self.complete()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test build image."""
import json
import os
from pathlib import Path
from random import choices
from string import ascii_letters
from typing import Tuple

import docker
from aea.cli.utils.config import get_default_author_from_cli_config
from aea.configurations.data_types import PackageId, PackageType, PublicId
from aea.configurations.loader import ConfigLoader
from aea.helpers.io import open_file

from autonomy.configurations.base import Service
from autonomy.constants import DEFAULT_DOCKER_IMAGE_AUTHOR

from tests.conftest import get_file_from_tag, skip_docker_tests
from tests.test_autonomy.base import get_dummy_service_config
from tests.test_autonomy.test_cli.base import BaseCliTest


@skip_docker_tests
class TestBuildImage(BaseCliTest):
    """Test build image command."""

    cli_options: Tuple[str, ...] = ("build-image",)
    docker_api: docker.APIClient
    build_dir: Path
    package_hash: str
    package_id: PackageId

    def setup(self) -> None:
        """Setup class."""
        super().setup()

        self.docker_api = docker.APIClient()
        self.package_id = PackageId(
            package_type=PackageType.AGENT,
            public_id=PublicId(author="valory", name="test_abci", version="0.1.0"),
        )

        packages_json = json.loads(get_file_from_tag("packages/packages.json"))
        package_hash = packages_json["dev"][self.package_id.to_uri_path]
        self.package_id = self.package_id.with_hash(package_hash=package_hash)

        os.chdir(self.t)

        service_config, *_ = get_dummy_service_config()
        service_config["agent"] = str(self.package_id.public_id)
        service_file = self.t / "service.yaml"

        with open_file(service_file, "w+") as fp:
            service = Service.from_json(service_config)
            ConfigLoader(Service.schema, Service).dump(service, fp)

    def test_build_prod(
        self,
    ) -> None:
        """Test prod build."""

        result = self.run_cli()

        assert result.exit_code == 0, result.output
        assert (
            len(
                self.docker_api.images(
                    name=f"{get_default_author_from_cli_config() or DEFAULT_DOCKER_IMAGE_AUTHOR}/oar-{self.package_id.name}:{self.package_id.package_hash}"
                )
            )
            == 1
        )

    def test_build_dev(
        self,
    ) -> None:
        """Test prod build."""

        result = self.run_cli(("--dev",))

        assert result.exit_code == 0, result.output
        assert (
            len(
                self.docker_api.images(
                    name=f"{get_default_author_from_cli_config() or DEFAULT_DOCKER_IMAGE_AUTHOR}/oar-{self.package_id.name}:dev"
                )
            )
            == 1
        )

    def test_build_version(
        self,
    ) -> None:
        """Test prod build."""

        test_version = "".join(choices(ascii_letters, k=6))  # nosec
        result = self.run_cli(("--version", test_version))

        assert result.exit_code == 0, result.output
        assert (
            len(
                self.docker_api.images(
                    name=f"{get_default_author_from_cli_config() or DEFAULT_DOCKER_IMAGE_AUTHOR}/oar-{self.package_id.name}:{test_version}"
                )
            )
            == 1
        )

    @skip_docker_tests
    def test_image_author_flag(self) -> None:
        """Test image_author flag."""
        test_version = "".join(choices(ascii_letters, k=6))  # nosec
        default_author = (
            get_default_author_from_cli_config() or DEFAULT_DOCKER_IMAGE_AUTHOR
        )
        result = self.run_cli(
            (
                "--version",
                test_version,
            )
        )
        assert result.exit_code == 0, result.output
        assert self.docker_api.images(
            name=f"{default_author}/oar-{self.package_id.name}:{test_version}"
        )

        test_version = "".join(choices(ascii_letters, k=6))  # nosec
        image_author = "test_author"
        result = self.run_cli(
            (
                "--version",
                test_version,
                "--image-author",
                image_author,
            )
        )
        assert result.exit_code == 0, result.output
        assert self.docker_api.images(
            name=f"{image_author}/oar-{self.package_id.name}:{test_version}"
        )

        test_version = "".join(choices(ascii_letters, k=6))  # nosec
        image_author = "0bad*author"
        result = self.run_cli(
            (
                "--version",
                test_version,
                "--image-author",
                image_author,
            )
        )
        assert result.exit_code == 1, result.output
        assert f"Value {image_author} does not match the regular expression " in str(
            result
        )


class TestBuildImageFailures(BaseCliTest):
    """Test build image command."""

    cli_options: Tuple[str, ...] = ("build-image",)
    docker_api: docker.APIClient
    build_dir: Path
    package_hash: str

    def test_service_file_missing(
        self,
    ) -> None:
        """Test prod build."""

        result = self.run_cli()

        assert result.exit_code == 1, result.output
        assert "No such file or directory: " in result.output
        assert "service.yaml" in result.output

    @skip_docker_tests
    def test_image_build_fail(self) -> None:
        """Test prod build."""

        result = self.run_cli(
            commands=(
                "valory/agent:bafybeihyasfforsfualp6jnhh2jj7nreqmws2ygyfnh4p3idmfkm5yxu11",
            )
        )

        assert result.exit_code == 1, result.output
        assert "Error occured while downloading agent" in result.output



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the data classes for the Hello World ABCI application."""

from abc import ABC
from enum import Enum
from typing import Dict, List, Optional, Tuple, Type, cast

from packages.valory.skills.abstract_round_abci.base import (
    AbciApp,
    AbciAppTransitionFunction,
    AbstractRound,
    AppState,
    BaseSynchronizedData,
    CollectDifferentUntilAllRound,
    CollectSameUntilAllRound,
    CollectSameUntilThresholdRound,
    get_name,
)
from packages.valory.skills.hello_world_abci.payloads import (
    CollectRandomnessPayload,
    PrintMessagePayload,
    RegistrationPayload,
    ResetPayload,
    SelectKeeperPayload,
)


class Event(Enum):
    """Event enumeration for the Hello World ABCI demo."""

    DONE = "done"
    ROUND_TIMEOUT = "round_timeout"
    NO_MAJORITY = "no_majority"
    RESET_TIMEOUT = "reset_timeout"


class SynchronizedData(
    BaseSynchronizedData
):  # pylint: disable=too-many-instance-attributes
    """
    Class to represent the synchronized data.

    This state is replicated by the Tendermint application.
    """

    @property
    def printed_messages(self) -> List[str]:
        """Get the printed messages list."""

        return cast(
            List[str],
            self.db.get_strict("printed_messages"),
        )


class HelloWorldABCIAbstractRound(AbstractRound, ABC):
    """Abstract round for the Hello World ABCI skill."""

    synchronized_data_class: Type[BaseSynchronizedData] = SynchronizedData

    @property
    def synchronized_data(self) -> SynchronizedData:
        """Return the synchronized data."""
        return cast(SynchronizedData, self._synchronized_data)


class RegistrationRound(CollectSameUntilAllRound, HelloWorldABCIAbstractRound):
    """A round in which the agents get registered"""

    payload_class = RegistrationPayload

    def end_block(self) -> Optional[Tuple[BaseSynchronizedData, Event]]:
        """Process the end of the block."""

        if self.collection_threshold_reached:
            synchronized_data = self.synchronized_data.update(
                participants=tuple(sorted(self.collection)),
                synchronized_data_class=SynchronizedData,
            )
            return synchronized_data, Event.DONE
        return None


class CollectRandomnessRound(
    CollectSameUntilThresholdRound, HelloWorldABCIAbstractRound
):
    """A round for collecting randomness"""

    payload_class = CollectRandomnessPayload
    synchronized_data_class = SynchronizedData
    done_event = Event.DONE
    no_majority_event = Event.NO_MAJORITY
    collection_key = get_name(SynchronizedData.participant_to_randomness)
    selection_key = get_name(SynchronizedData.most_voted_randomness)


class SelectKeeperRound(CollectSameUntilThresholdRound, HelloWorldABCIAbstractRound):
    """A round in a which keeper is selected"""

    payload_class = SelectKeeperPayload
    synchronized_data_class = SynchronizedData
    done_event = Event.DONE
    no_majority_event = Event.NO_MAJORITY
    collection_key = get_name(SynchronizedData.participant_to_selection)
    selection_key = get_name(SynchronizedData.most_voted_keeper_address)


class PrintMessageRound(CollectDifferentUntilAllRound, HelloWorldABCIAbstractRound):
    """A round in which the keeper prints the message"""

    payload_class = PrintMessagePayload

    def end_block(self) -> Optional[Tuple[BaseSynchronizedData, Event]]:
        """Process the end of the block."""
        if self.collection_threshold_reached:
            synchronized_data = self.synchronized_data.update(
                participants=tuple(sorted(self.collection)),
                printed_messages=sorted(
                    [
                        cast(PrintMessagePayload, payload).message
                        for payload in self.collection.values()
                    ]
                ),
                synchronized_data_class=SynchronizedData,
            )
            return synchronized_data, Event.DONE
        return None


class ResetAndPauseRound(CollectSameUntilThresholdRound, HelloWorldABCIAbstractRound):
    """This class represents the base reset round."""

    payload_class = ResetPayload

    def end_block(self) -> Optional[Tuple[BaseSynchronizedData, Event]]:
        """Process the end of the block."""
        if self.threshold_reached:
            return self.synchronized_data.create(), Event.DONE
        if not self.is_majority_possible(
            self.collection, self.synchronized_data.nb_participants
        ):
            return self.synchronized_data, Event.NO_MAJORITY
        return None


class HelloWorldAbciApp(AbciApp[Event]):
    """HelloWorldAbciApp

    Initial round: RegistrationRound

    Initial states: {RegistrationRound}

    Transition states:
        0. RegistrationRound
            - done: 1.
        1. CollectRandomnessRound
            - done: 2.
            - no majority: 1.
            - round timeout: 1.
        2. SelectKeeperRound
            - done: 3.
            - no majority: 0.
            - round timeout: 0.
        3. PrintMessageRound
            - done: 4.
            - round timeout: 0.
        4. ResetAndPauseRound
            - done: 1.
            - no majority: 0.
            - reset timeout: 0.

    Final states: {}

    Timeouts:
        round timeout: 30.0
        reset timeout: 30.0
    """

    initial_round_cls: AppState = RegistrationRound
    transition_function: AbciAppTransitionFunction = {
        RegistrationRound: {
            Event.DONE: CollectRandomnessRound,
        },
        CollectRandomnessRound: {
            Event.DONE: SelectKeeperRound,
            Event.NO_MAJORITY: CollectRandomnessRound,
            Event.ROUND_TIMEOUT: CollectRandomnessRound,
        },
        SelectKeeperRound: {
            Event.DONE: PrintMessageRound,
            Event.NO_MAJORITY: RegistrationRound,
            Event.ROUND_TIMEOUT: RegistrationRound,
        },
        PrintMessageRound: {
            Event.DONE: ResetAndPauseRound,
            Event.ROUND_TIMEOUT: RegistrationRound,
        },
        ResetAndPauseRound: {
            Event.DONE: CollectRandomnessRound,
            Event.NO_MAJORITY: RegistrationRound,
            Event.RESET_TIMEOUT: RegistrationRound,
        },
    }
    event_to_timeout: Dict[Event, float] = {
        Event.ROUND_TIMEOUT: 30.0,
        Event.RESET_TIMEOUT: 30.0,
    }



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Base test."""

import itertools
from pathlib import Path
from typing import List

import _strptime  # noqa  # pylint: disable=unsed-import
import pytest
from aea.components.base import load_aea_package
from aea.configurations.constants import PACKAGES, PACKAGE_TYPE_TO_CONFIG_FILE
from aea.configurations.data_types import PackageType
from aea.configurations.loader import ConfigLoader
from aea.helpers.io import open_file
from aea_test_autonomy.configurations import (
    DEFAULT_IMAGE_VERSION as PLUGIN_DEFAULT_IMAGE_VERSION,
)
from aea_test_autonomy.configurations import (
    TENDERMINT_IMAGE_NAME as PLUGIN_TENDERMINT_IMAGE_NAME,
)
from aea_test_autonomy.configurations import (
    TENDERMINT_IMAGE_VERSION as PLUGIN_TENDERMINT_IMAGE_VERSION,
)

import autonomy
from autonomy.configurations.base import PACKAGE_TYPE_TO_CONFIG_CLASS
from autonomy.constants import (
    DEFAULT_IMAGE_VERSION,
    TENDERMINT_IMAGE_NAME,
    TENDERMINT_IMAGE_VERSION,
)

from tests.conftest import ROOT_DIR


def get_test_files(package_type: PackageType) -> List[Path]:
    """Returns the list of components to be tested."""
    return list(
        (ROOT_DIR / PACKAGES).glob(
            f"*/{package_type.to_plural()}/*/{PACKAGE_TYPE_TO_CONFIG_FILE[package_type.value]}"
        )
    )


def test_version() -> None:
    """Test the version."""
    assert autonomy.__version__ == "0.10.5.post2"


@pytest.mark.parametrize(
    "component_type,config_file_path",
    itertools.chain.from_iterable(
        [
            itertools.zip_longest([], files, fillvalue=component_type)
            for files, component_type in [
                (get_test_files(PackageType.PROTOCOL), PackageType.PROTOCOL),
                (get_test_files(PackageType.CONTRACT), PackageType.CONTRACT),
                (get_test_files(PackageType.CONNECTION), PackageType.CONNECTION),
                (get_test_files(PackageType.SKILL), PackageType.SKILL),
                (get_test_files(PackageType.SERVICE), PackageType.SERVICE),
            ]
        ]
    ),
)
def test_load_all_packages(component_type: PackageType, config_file_path: str) -> None:
    """Load all AEA component packages."""

    configuration_loader = ConfigLoader.from_configuration_type(
        component_type, PACKAGE_TYPE_TO_CONFIG_CLASS
    )
    with open_file(config_file_path) as fp:
        configuration_object = configuration_loader.load(fp)
        if component_type != PackageType.SERVICE:
            # we don't need to check sub packages for service components
            # so we perform `load_aea_package` only for none service components
            directory = Path(config_file_path).parent
            configuration_object.directory = directory
            load_aea_package(configuration_object)


def test_synced_with_plugins() -> None:
    """Test plugin and autonomy constants are synced."""
    assert DEFAULT_IMAGE_VERSION == PLUGIN_DEFAULT_IMAGE_VERSION
    assert TENDERMINT_IMAGE_NAME == PLUGIN_TENDERMINT_IMAGE_NAME
    assert TENDERMINT_IMAGE_VERSION == PLUGIN_TENDERMINT_IMAGE_VERSION



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test the utils.py module of the skill."""

from collections import defaultdict
from string import printable
from typing import Any, Dict, List, Tuple, Type
from unittest import mock

import pytest
from hypothesis import assume, given, settings
from hypothesis import strategies as st

from packages.valory.skills.abstract_round_abci.tests.conftest import profile_name
from packages.valory.skills.abstract_round_abci.utils import (
    DEFAULT_TENDERMINT_P2P_PORT,
    MAX_UINT64,
    VerifyDrand,
    consensus_threshold,
    filter_negative,
    get_data_from_nested_dict,
    get_value_with_type,
    is_json_serializable,
    is_primitive_or_none,
    parse_tendermint_p2p_url,
)


settings.load_profile(profile_name)


# pylint: skip-file


DRAND_PUBLIC_KEY: str = "868f005eb8e6e4ca0a47c8a77ceaa5309a47978a7c71bc5cce96366b5d7a569937c529eeda66c7293784a9402801af31"

DRAND_VALUE = {
    "round": 1416669,
    "randomness": "f6be4bf1fa229f22340c1a5b258f809ac4af558200775a67dacb05f0cb258a11",
    "signature": (
        "b44d00516f46da3a503f9559a634869b6dc2e5d839e46ec61a090e3032172954929a5"
        "d9bd7197d7739fe55db770543c71182562bd0ad20922eb4fe6b8a1062ed21df3b68de"
        "44694eb4f20b35262fa9d63aa80ad3f6172dd4d33a663f21179604"
    ),
    "previous_signature": (
        "903c60a4b937a804001032499a855025573040cb86017c38e2b1c3725286756ce8f33"
        "61188789c17336beaf3f9dbf84b0ad3c86add187987a9a0685bc5a303e37b008fba8c"
        "44f02a416480dd117a3ff8b8075b1b7362c58af195573623187463"
    ),
}


class TestVerifyDrand:
    """Test DrandVerify."""

    drand_check: VerifyDrand

    def setup(
        self,
    ) -> None:
        """Setup test."""
        self.drand_check = VerifyDrand()

    def test_verify(
        self,
    ) -> None:
        """Test verify method."""

        result, error = self.drand_check.verify(DRAND_VALUE, DRAND_PUBLIC_KEY)
        assert result
        assert error is None

    def test_verify_fails(
        self,
    ) -> None:
        """Test verify method."""

        drand_value = DRAND_VALUE.copy()
        del drand_value["randomness"]
        result, error = self.drand_check.verify(drand_value, DRAND_PUBLIC_KEY)
        assert not result
        assert error == "DRAND dict is missing value for 'randomness'"

        drand_value = DRAND_VALUE.copy()
        drand_value["randomness"] = "".join(
            list(drand_value["randomness"])[:-1] + ["0"]  # type: ignore
        )
        result, error = self.drand_check.verify(drand_value, DRAND_PUBLIC_KEY)
        assert not result
        assert error == "Failed randomness hash check."

        drand_value = DRAND_VALUE.copy()
        with mock.patch.object(
            self.drand_check, "_verify_signature", return_value=False
        ):
            result, error = self.drand_check.verify(drand_value, DRAND_PUBLIC_KEY)

        assert not result
        assert error == "Failed bls.Verify check."

    @pytest.mark.parametrize("value", (-1, MAX_UINT64 + 1))
    def test_negative_and_overflow(self, value: int) -> None:
        """Test verify method."""
        with pytest.raises(ValueError):
            self.drand_check._int_to_bytes_big(value)


@given(st.integers(min_value=0, max_value=MAX_UINT64))
def test_verify_int_to_bytes_big_fuzz(integer: int) -> None:
    """Test VerifyDrand."""

    VerifyDrand._int_to_bytes_big(integer)


@pytest.mark.parametrize("integer", [-1, MAX_UINT64 + 1])
def test_verify_int_to_bytes_big_raises(integer: int) -> None:
    """Test VerifyDrand._int_to_bytes_big"""

    expected = "VerifyDrand can only handle positive numbers representable with 8 bytes"
    with pytest.raises(ValueError, match=expected):
        VerifyDrand._int_to_bytes_big(integer)


@given(st.binary())
def test_verify_randomness_hash_fuzz(input_bytes: bytes) -> None:
    """Test VerifyDrand._verify_randomness_hash"""

    VerifyDrand._verify_randomness_hash(input_bytes, input_bytes)


@given(
    st.lists(st.text(), min_size=1, max_size=50),
    st.binary(),
    st.characters(),
)
def test_get_data_from_nested_dict(
    nested_keys: List[str], final_value: bytes, separator: str
) -> None:
    """Test `get_data_from_nested_dict`"""
    assume(not any(separator in key for key in nested_keys))

    def create_nested_dict() -> defaultdict:
        """Recursively create a nested dict of arbitrary size."""
        return defaultdict(create_nested_dict)

    nested_dict = create_nested_dict()
    key_access = (f"[nested_keys[{i}]]" for i in range(len(nested_keys)))
    expression = "nested_dict" + "".join(key_access)
    expression += " = final_value"
    exec(expression)  # nosec

    serialized_keys = separator.join(nested_keys)
    actual = get_data_from_nested_dict(nested_dict, serialized_keys, separator)
    assert actual == final_value


@pytest.mark.parametrize(
    "type_name, type_, value",
    (
        ("str", str, "1"),
        ("int", int, 1),
        ("float", float, 1.1),
        ("dict", dict, {1: 1}),
        ("list", list, [1]),
        ("non_existent", None, 1),
    ),
)
def test_get_value_with_type(type_name: str, type_: Type, value: Any) -> None:
    """Test `get_value_with_type`"""
    if type_ is None:
        with pytest.raises(
            AttributeError, match=f"module 'builtins' has no attribute '{type_name}'"
        ):
            get_value_with_type(value, type_name)
        return

    actual = get_value_with_type(value, type_name)
    assert type(actual) == type_
    assert actual == value


@pytest.mark.parametrize(
    ("url", "expected_output"),
    (
        ("localhost", ("localhost", DEFAULT_TENDERMINT_P2P_PORT)),
        ("localhost:80", ("localhost", 80)),
        ("some.random.host:80", ("some.random.host", 80)),
        ("1.1.1.1", ("1.1.1.1", DEFAULT_TENDERMINT_P2P_PORT)),
        ("1.1.1.1:80", ("1.1.1.1", 80)),
    ),
)
def test_parse_tendermint_p2p_url(url: str, expected_output: Tuple[str, int]) -> None:
    """Test `parse_tendermint_p2p_url` method."""

    assert parse_tendermint_p2p_url(url=url) == expected_output


@given(
    st.one_of(st.none(), st.integers(), st.floats(), st.text(), st.booleans()),
    st.one_of(
        st.nothing(),
        st.frozensets(st.integers()),
        st.sets(st.integers()),
        st.lists(st.integers()),
        st.dictionaries(st.integers(), st.integers()),
        st.dates(),
        st.complex_numbers(),
        st.just(object()),
    ),
)
def test_is_primitive_or_none(valid_obj: Any, invalid_obj: Any) -> None:
    """Test `is_primitive_or_none`."""
    assert is_primitive_or_none(valid_obj)
    assert not is_primitive_or_none(invalid_obj)


@given(
    st.recursive(
        st.none() | st.booleans() | st.floats() | st.text(printable),
        lambda children: st.lists(children)
        | st.dictionaries(st.text(printable), children),
    ),
    st.one_of(
        st.nothing(),
        st.frozensets(st.integers()),
        st.sets(st.integers()),
        st.dates(),
        st.complex_numbers(),
        st.just(object()),
    ),
)
def test_is_json_serializable(valid_obj: Any, invalid_obj: Any) -> None:
    """Test `is_json_serializable`."""
    assert is_json_serializable(valid_obj)
    assert not is_json_serializable(invalid_obj)


@given(
    positive=st.dictionaries(st.text(), st.integers(min_value=0)),
    negative=st.dictionaries(st.text(), st.integers(max_value=-1)),
)
def test_filter_negative(positive: Dict[str, int], negative: Dict[str, int]) -> None:
    """Test `filter_negative`."""
    assert len(tuple(filter_negative(positive))) == 0
    assert set(filter_negative(negative)) == set(negative.keys())


@pytest.mark.parametrize(
    "nb, threshold",
    ((1, 1), (2, 2), (3, 3), (4, 3), (5, 4), (6, 5), (100, 67), (300, 201)),
)
def test_consensus_threshold(nb: int, threshold: int) -> None:
    """Test `consensus_threshold`."""
    assert consensus_threshold(nb) == threshold



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
# pylint: disable=protected-access,attribute-defined-outside-init

"""Tests for ipfs connection."""
import asyncio
import os
import platform
import tempfile
from typing import Dict, List, Optional
from unittest import mock
from unittest.mock import MagicMock

import pytest
from aea.configurations.base import ConnectionConfig
from aea.connections.base import ConnectionStates
from aea.mail.base import Envelope
from aea_cli_ipfs.exceptions import DownloadError
from aea_cli_ipfs.ipfs_utils import IPFSTool
from aea_test_autonomy.fixture_helpers import (  # noqa: F401; pylint: disable=unused-import
    LOCAL_IPFS,
    ipfs_daemon,
    use_ipfs_daemon,
)

from packages.valory.connections.ipfs.connection import (
    IpfsConnection,
    IpfsDialogues,
    PUBLIC_ID,
)
from packages.valory.protocols.ipfs import IpfsMessage


ANY_SKILL = "skill/any:0.1.0"


@use_ipfs_daemon
class TestIpfsConnection:
    """Tests for IpfsConnection"""

    def setup(self) -> None:
        """Set up the tests."""
        configuration = ConnectionConfig(
            ipfs_domain=LOCAL_IPFS,
            connection_id=IpfsConnection.connection_id,
        )
        self.connection = IpfsConnection(
            configuration=configuration,
            data_dir=MagicMock(),
        )
        self.dummy_msg = IpfsMessage(performative=IpfsMessage.Performative.GET_FILES)  # type: ignore
        self.dummy_envelope = Envelope(
            to=str(PUBLIC_ID), sender=ANY_SKILL, message=self.dummy_msg
        )

    @pytest.mark.asyncio
    async def test_connect(self) -> None:
        """Test connect"""
        await self.connection.connect()
        assert self.connection.response_envelopes is not None
        assert self.connection.state == ConnectionStates.connected

    @pytest.mark.asyncio
    async def test_disconnect(self) -> None:
        """Test disconnect"""
        await self.connection.connect()
        await self.connection.disconnect()
        assert self.connection.state == ConnectionStates.disconnected
        with pytest.raises(ValueError):
            self.connection.response_envelopes  # pylint: disable=pointless-statement

    @pytest.mark.parametrize(
        ("performative", "expected_error"),
        [
            (IpfsMessage.Performative.GET_FILES, False),
            (IpfsMessage.Performative.STORE_FILES, False),
            (IpfsMessage.Performative.IPFS_HASH, True),
            (IpfsMessage.Performative.FILES, True),
        ],
    )
    def test_handle_envelope(
        self,
        performative: IpfsMessage.Performative,
        expected_error: bool,
    ) -> None:
        """Tests for _handle_envelope."""
        dummy_msg = IpfsMessage(performative=performative)
        envelope = Envelope(to=str(PUBLIC_ID), sender=ANY_SKILL, message=dummy_msg)
        with mock.patch.object(
            IpfsConnection,
            "run_async",
            return_value=MagicMock(),
        ), mock.patch.object(
            self.connection.logger,
            "error",
        ) as mock_logger:
            self.connection._handle_envelope(envelope)
            if expected_error:
                mock_logger.assert_called_with(
                    f"Performative `{performative.value}` is not supported."
                )

    @pytest.mark.asyncio
    async def test_send(self) -> None:
        """Test send."""
        await self.connection.connect()
        with mock.patch.object(
            IpfsConnection,
            "_handle_envelope",
            return_value=MagicMock(add_done_callback=lambda *_: _),
        ):
            assert len(self.connection.task_to_request) == 0
            await self.connection.send(self.dummy_envelope)
            assert len(self.connection.task_to_request) == 1

    @pytest.mark.asyncio
    async def test_receive(self) -> None:
        """Tests receive."""
        await self.connection.connect()
        expected_envelope = self.dummy_envelope
        self.connection.response_envelopes.put_nowait(expected_envelope)
        actual_envelope = await self.connection.receive()

        assert expected_envelope == actual_envelope

    @pytest.mark.asyncio
    async def test_handle_done_task(self) -> None:
        """Test _handle_done_tasks."""
        await self.connection.connect()
        assert self.connection.response_envelopes.qsize() == 0
        dummy_task = MagicMock(result=lambda: self.dummy_msg)
        dummy_request = MagicMock(to=ANY_SKILL, sender=str(PUBLIC_ID), context=None)
        self.connection.task_to_request[dummy_task] = dummy_request
        self.connection._handle_done_task(dummy_task)
        assert self.connection.response_envelopes.qsize() == 1

    def test_handle_error(self) -> None:
        """Test handle_error."""
        message = self.connection._handle_error(
            reason="dummy reason",
            dialogue=MagicMock(),
        )
        assert message is not None

    @pytest.mark.asyncio
    async def test_run_async(self) -> None:
        """Test run async."""
        dummy_log = "dummy operation is being performed"

        def dummy_operation() -> None:
            """A dummy handler."""
            self.connection.logger.info(dummy_log)

        await self.connection.connect()
        with mock.patch.object(
            self.connection.logger,
            "info",
        ) as mock_logger:
            task = self.connection.run_async(dummy_operation)
            assert task is not None

            # give some time for the task to be scheduled and run
            await asyncio.sleep(1)
            mock_logger.assert_called_with(dummy_log)

    @pytest.mark.parametrize(
        ("files", "expected_log", "log_level"),
        [
            ({}, "No files were present.", "error"),
            (
                {"dummy_filename": "dummy_content"},
                "Successfully stored files with hash: QmYn1qHyFMDdVxYcwseLBdooLAyc3orNBH8vvj4iPTXd4R.",
                "debug",
            ),
            (
                {
                    "dummy_dir/dummy_filename1": "dummy_content",
                    "dummy_dir/dummy_filename2": "dummy_content",
                },
                "Successfully stored files with hash: QmRP7wK1NZKwno9CFxQUeFaJip5eaaLBo1v8WFtyECgLCz.",
                "debug",
            ),
            (
                {
                    "dummy_dir1/dummy_filename1": "dummy_content",
                    "dummy_dir2/dummy_filename2": "dummy_content",
                },
                "If you want to send multiple files as a single dir, "
                "make sure the their path matches to one directory only.",
                "info",
            ),
        ],
    )
    def test_handle_store_files(
        self, files: Dict[str, str], expected_log: str, log_level: str
    ) -> None:
        """Test _handle_store_files."""
        with mock.patch.object(
            self.connection.logger,
            log_level,
        ) as mock_logger:
            message = IpfsMessage(
                performative=IpfsMessage.Performative.STORE_FILES, files=files  # type: ignore
            )
            dialogue = MagicMock()
            message = self.connection._handle_store_files(message, dialogue)
            assert message is not None
            mock_logger.assert_called_with(expected_log)

    @pytest.mark.skipif(
        platform.system() == "Windows",
        reason="PermissionError: [Errno 13] Permission denied",
    )
    @pytest.mark.parametrize(
        ("extra_files", "download_side_effect", "is_dir"),
        [
            ([], None, False),
            (["dummy_unexpected_extra_file"], None, False),
            ([], DownloadError, False),
            ([], None, True),
        ],
    )
    def test_handle_get_files(
        self,
        extra_files: List[str],
        download_side_effect: Optional[Exception],
        is_dir: bool,
    ) -> None:
        """Test _handle_get_files"""
        with tempfile.NamedTemporaryFile() as tmp_file, mock.patch.object(
            os,
            "listdir",
        ) as mock_listdir, mock.patch.object(
            IPFSTool, "download", side_effect=download_side_effect
        ):
            listdir_value = [extra_files + [tmp_file.name]]
            if is_dir:
                listdir_value = [["/tmp"]] + listdir_value  # nosec
            mock_listdir.side_effect = listdir_value

            tmp_file.write(b"dummy_data")
            tmp_file.flush()

            _, ipfs_hash, _ = self.connection.ipfs_tool.add(tmp_file.name)
            message = IpfsMessage(
                performative=IpfsMessage.Performative.GET_FILES, ipfs_hash=ipfs_hash  # type: ignore
            )
            dialogue = MagicMock()
            message = self.connection._handle_get_files(message, dialogue)
            assert message is not None

    def test_ipfs_dialogue(self) -> None:  # pylint: disable=no-self-use
        """Test 'IpfsDialogues' creation."""
        dialogues = IpfsDialogues(connection_id=str(PUBLIC_ID))
        dialogues.create(
            counterparty=ANY_SKILL,
            performative=IpfsMessage.Performative.GET_FILES,
        )



# flake8: noqa
# type: ignore
# pylint: skip-file
click_context  # unused variable (autonomy/analyse/__init__.py:30)
cmd1  # unused function (autonomy/analyse/cli.py:25)
build_deployment  # unused function (autonomy/cli/deploy.py:38)
dev_mode  # unused variable (autonomy/cli/deploy.py:67)
schema_dir  # unused variable (autonomy/configurations/base.py:32)
_validator  # unused attribute (autonomy/deploy/base.py:111)
_validator  # unused attribute (autonomy/deploy/base.py:115)
agent_spec  # unused attribute (autonomy/deploy/base.py:330)
old_wd  # unused variable (autonomy/deploy/base.py:402)
build_docker_compose_yml  # unused function (autonomy/deploy/generators/docker_compose/base.py:52)
HARDHAT_NODE_TEMPLATE  # unused variable (autonomy/deploy/generators/docker_compose/templates.py:51)
HARDHAT_TEMPLATE  # unused variable (autonomy/deploy/generators/kubernetes/templates.py:25)
run_agent  # unused function (autonomy/cli/replay.py:44)
run_tendermint  # unused function (autonomy/cli/replay.py:78)
run_until_interruption  # unused method (autonomy/replay/tendermint.py:178)
hard_reset  # unused function (autonomy/replay/tendermint.py:195)
status  # unused function (autonomy/replay/tendermint.py:215)
broadcast_tx_sync  # unused function (autonomy/replay/tendermint.py:238)
tx  # unused function (autonomy/replay/tendermint.py:244)
LOCAL  # unused variable (autonomy/analyse/benchmark/aggregate.py:35)
CONSENSUS  # unused variable (autonomy/analyse/benchmark/aggregate.py:36)
TOTAL  # unused variable (autonomy/analyse/benchmark/aggregate.py:37)
benchmark  # unused function (autonomy/cli/analyse.py:35)
docstrings  # unused function (autonomy/cli/analyse.py:49)
parse_logs  # unused function (autonomy/cli/analyse.py:74)
HARDHAT_IMAGE_NAME  # unused variable (autonomy/constants.py:25)
is_transition_func_total  # unused method (autonomy/analyse/abci/app_spec.py:119)
get_transitions  # unused method (autonomy/analyse/abci/app_spec.py:134)
abci_app_specs  # unused function (autonomy/cli/analyse.py:52)
check_abci_app_specs  # unused function (autonomy/cli/analyse.py:79)
run_handler_check  # unused function (autonomy/cli/analyse.py:159)
dump_json  # unused method (autonomy/analyse/abci/app_spec.py:159)
dump_yaml  # unused method (autonomy/analyse/abci/app_spec.py:163)
dump_mermaid  # unused method (autonomy/analyse/abci/app_spec.py:167)
from_json  # unused method  (autonomy/configurations/base.py:100)
SERVICES  # unused variable  (autonomy/configurations/constants.py:30)
_directory  # unused attribute (autonomy/cli/hash.py:63)
generate_all  # unused function (autonomy/cli/hash.py:72)
hash_file  # unused function (autonomy/cli/hash.py:94)
_create_or_update_from_json  # unused method (autonomy/configurations/base.py:175)
help  # unused attribute (autonomy/cli/core.py:41)
DEFAULT_ABCI_BUILD_DIR  # unused variable (autonomy/deploy/constants.py:105)
build_images  # unused function (autonomy/cli/deploy.py:187)
execinfo  # unused variable (autonomy/data/Dockerfiles/tendermint/app.py:100)
get_params  # unused function (autonomy/data/Dockerfiles/tendermint/app.py:142)
update_params  # unused function (autonomy/data/Dockerfiles/tendermint/app.py:155)
gentle_reset  # unused function (autonomy/data/Dockerfiles/tendermint/app.py:187)
app_hash  # unused function (autonomy/data/Dockerfiles/tendermint/app.py:197)
handle_notfound  # unused function (autonomy/data/Dockerfiles/tendermint/app.py:236)
handle_server_error  # unused function (autonomy/data/Dockerfiles/tendermint/app.py:242)
create_server  # unused function (autonomy/data/Dockerfiles/tendermint/app.py:251)
call_vote  # unused function (autonomy/data/Dockerfiles/dev/watcher.py:54)
wait_for_votes  # unused function (autonomy/data/Dockerfiles/dev/watcher.py:67)
on_any_event  # unused method (autonomy/data/Dockerfiles/dev/watcher.py:166)
tendermint_health_check  # unused function (autonomy/test_tools/utils.py:30)
DEFAULT_ACN_CONFIG  # unused variable (autonomy/test_tools/docker/acn_node.py:35)
ACNNodeDockerImage  # unused class (autonomy/test_tools/docker/acn_node.py:45)
SAFE_CONTRACT  # unused variable (autonomy/test_tools/docker/amm_net.py:41)
ROUTER_CONTRACT  # unused variable (autonomy/test_tools/docker/amm_net.py:42)
DEFAULT_CALLBACK_HANDLER  # unused variable (autonomy/test_tools/docker/amm_net.py:43)
PROXY_FACTORY_CONTRACT  # unused variable (autonomy/test_tools/docker/amm_net.py:44)
MULTISEND_CONTRACT  # unused variable (autonomy/test_tools/docker/amm_net.py:45)
MULTISEND_CALL_ONLY_CONTRACT  # unused variable (autonomy/test_tools/docker/amm_net.py:46)
AMMNetDockerImage  # unused class (autonomy/test_tools/docker/amm_net.py:49)
launch_image  # unused function (autonomy/test_tools/docker/base.py:173)
launch_many_containers  # unused function (autonomy/test_tools/docker/base.py:191)
_.setup_class  # unused method (autonomy/test_tools/docker/base.py:224)
_.teardown_class  # unused method (autonomy/test_tools/docker/base.py:246)
DEFAULT_GANACHE_ADDR  # unused variable (autonomy/test_tools/docker/ganache.py:33)
DEFAULT_GANACHE_PORT  # unused variable (autonomy/test_tools/docker/ganache.py:34)
DEFAULT_GANACHE_CHAIN_ID  # unused variable (autonomy/test_tools/docker/ganache.py:35)
GanacheDockerImage  # unused class (autonomy/test_tools/docker/ganache.py:38)
SAFE_CONTRACT  # unused variable (autonomy/test_tools/docker/gnosis_safe_net.py:44)
DEFAULT_CALLBACK_HANDLER  # unused variable (autonomy/test_tools/docker/gnosis_safe_net.py:45)
PROXY_FACTORY_CONTRACT  # unused variable (autonomy/test_tools/docker/gnosis_safe_net.py:46)
MULTISEND_CONTRACT  # unused variable (autonomy/test_tools/docker/gnosis_safe_net.py:47)
MULTISEND_CALL_ONLY_CONTRACT  # unused variable (autonomy/test_tools/docker/gnosis_safe_net.py:48)
GnosisSafeNetDockerImage  # unused class (autonomy/test_tools/docker/gnosis_safe_net.py:51)
FlaskTendermintDockerImage  # unused class (autonomy/test_tools/docker/tendermint.py:119)
_.health_check  # unused method (autonomy/test_tools/docker/tendermint.py:371)
ETHEREUM_ENCRYPTED_KEYS  # unused variable (autonomy/test_tools/configurations.py:46)
ETHEREUM_ENCRYPTION_PASSWORD  # unused variable (autonomy/test_tools/configurations.py:47)
GANACHE_CONFIGURATION  # unused variable (autonomy/test_tools/configurations.py:48)
HTTP_LOCALHOST  # unused variable (autonomy/test_tools/configurations.py:39)
KEY_PAIRS  # unused variable (autonomy/test_tools/configurations.py:62)
UseTendermint  # unused class (autonomy/test_tools/fixture_helpers.py:55)
_._start_tendermint  # unused method (autonomy/test_tools/fixture_helpers.py:62)
_.node_address  # unused property (autonomy/test_tools/fixture_helpers.py:81)
UseFlaskTendermintNode  # unused class (autonomy/test_tools/fixture_helpers.py:87)
_._start_tendermint  # unused method (autonomy/test_tools/fixture_helpers.py:91)
_.get_laddr  # unused method (autonomy/test_tools/fixture_helpers.py:120)
UseGnosisSafeHardHatNet  # unused class (autonomy/test_tools/fixture_helpers.py:129)
_._start_hardhat  # unused method (autonomy/test_tools/fixture_helpers.py:135)
gnosis_safe_hardhat_scope_function  # unused variable (autonomy/test_tools/fixture_helpers.py:138)
hardhat_port  # unused variable (autonomy/test_tools/fixture_helpers.py:138)
UseGanache  # unused class (autonomy/test_tools/fixture_helpers.py:144)
_._start_ganache  # unused method (autonomy/test_tools/fixture_helpers.py:150)
ganache  # unused variable (autonomy/test_tools/fixture_helpers.py:152)
UseACNNode  # unused class (autonomy/test_tools/fixture_helpers.py:163)
_._start_acn  # unused method (autonomy/test_tools/fixture_helpers.py:170)
_._acn_node_image  # unused attribute (autonomy/test_tools/fixture_helpers.py:174)
get_register_contract  # unused function (autonomy/test_tools/helpers/contracts.py:30)
GanacheForkDockerImage  # unused function (autonomy/test_tools/docker/ganache.py:111)
try_send  # unused function (/home/marcofavorito/workfolder/open-autonomy/autonomy/test_tools/helpers/base.py:68)
ANY_ADDRESS  # unused variable (/home/marcofavorito/workfolder/open-autonomy/autonomy/test_tools/configurations.py:144)
wait_for_condition  # unused function (/home/marcofavorito/workfolder/open-autonomy/autonomy/test_tools/helpers/async_utils.py:29)
BaseThreadedAsyncLoop  # unused class (/home/marcofavorito/workfolder/open-autonomy/autonomy/test_tools/helpers/async_utils.py:146)
thread  # unused variable (/home/marcofavorito/workfolder/open-autonomy/autonomy/test_tools/helpers/async_utils.py:151)
_.setup  # unused method (/home/marcofavorito/workfolder/open-autonomy/autonomy/test_tools/helpers/async_utils.py:153)
_.execute  # unused method (/home/marcofavorito/workfolder/open-autonomy/autonomy/test_tools/helpers/async_utils.py:158)
_.teardown  # unused method (/home/marcofavorito/workfolder/open-autonomy/autonomy/test_tools/helpers/async_utils.py:163)
_.get_command  # unused method (/home/marcofavorito/workfolder/open-autonomy/autonomy/test_tools/helpers/tendermint_utils.py:125)
capture_log  # unused variable (/home/marcofavorito/workfolder/open-autonomy/autonomy/test_tools/base_test_classes/agents.py:71)
cli_log_options  # unused variable (/home/marcofavorito/workfolder/open-autonomy/autonomy/test_tools/base_test_classes/agents.py:79)
is_terminating  # unused variable (/home/marcofavorito/workfolder/open-autonomy/autonomy/test_tools/base_test_classes/agents.py:283)
run_deployment  # unused function (autonomy/cli/deploy.py:150)
image_profile_flag  # unused function (autonomy/cli/utils/click_utils.py:30)
DATA_DIR  # unused variable (autonomy/data/__init__.py:25)
DOCKERFILES  # unused variable (autonomy/deploy/constants.py:61)
build_image  # unused function (autonomy/deploy/image.py:58)
build_deployment_command  # unused function (autonomy/cli/deploy.py:64)
run_deployment_from_token  # unused function (autonomy/cli/deploy.py:189)
DEFAULT_ACCOUNT  # unused variable (autonomy/test_tools/docker/registries.py:40)
COMPONENT_REGISTRY  # unused variable (autonomy/test_tools/docker/registries.py:41)
AGENT_REGISTRY  # unused variable (autonomy/test_tools/docker/registries.py:42)
REGISTRIES_MANAGER  # unused variable (autonomy/test_tools/docker/registries.py:43)
GNOSIS_SAFE_MULTISIG  # unused variable (autonomy/test_tools/docker/registries.py:44)
SERVICE_REGISTRY  # unused variable (autonomy/test_tools/docker/registries.py:45)
SERVICE_MANAGER  # unused variable (autonomy/test_tools/docker/registries.py:46)
SERVICE_MULTISIG  # unused variable (autonomy/test_tools/docker/registries.py:47)
RegistriesDockerImage  # unused class (autonomy/test_tools/docker/registries.py:53)
identity  # unused function (autonomy/test_tools/helpers/base.py:87)
skip_docker_tests  # unused variable (autonomy/test_tools/docker/base.py:40)
run_service_locally  # unused function (autonomy.cli.develop.run_service_locally)
lock_packages  # unused function (autonomy/cli/packages.py:33)
PathArgument  # unused class (autonomy/cli/utils/click_utils.py:91)
protocol  # unused function (autonomy/cli/develop.py:101)
local  # unused variable (autonomy/chain/config.py:71)
custom_chain  # unused variable (autonomy/chain/config.py:77)
goerli  # unused variable (autonomy/chain/config.py:83)
ethereum  # unused variable (autonomy/chain/config.py:89)
agent_registry  # unused variable (autonomy/chain/config.py:116)
service_manager  # unused variable (autonomy/chain/config.py:138)
gnosis_safe_proxy_factory  # unused variable (autonomy/chain/config.py:163)
COMPONENT_REGISTRY_ABI_FILENAME  # unused variable (autonomy/chain/constants.py:38)
AGENT_REGISTRY_ABI_FILENAME  # unused variable (autonomy/chain/constants.py:39)
REGISTRIES_MANAGER_ABI_FILENAME  # unused variable (autonomy/chain/constants.py:40)
SERVICE_MANAGER_ABI_FILENAME  # unused variable (autonomy/chain/constants.py:41)
GNOSIS_SAFE_MULTISIG_ADDRESS_LOCAL  # unused variable (autonomy/chain/constants.py:51)
GNOSIS_SAFE_MULTISIG_ADDRESS_ETHEREUM  # unused variable (autonomy/chain/constants.py:60)
GNOSIS_SAFE_MULTISIG_ADDRESS_GOERLI  # unused variable (autonomy/chain/constants.py:68)
GNOSIS_SAFE_MULTISIG_ADDRESS_CUSTOM  # unused variable (autonomy/chain/constants.py:76)
COMPONENT_REGISTRY_ADDRESS_LOCAL  # unused variable (autonomy/chain/constants.py:37)
AGENT_REGISTRY_ADDRESS_LOCAL  # unused variable (autonomy/chain/constants.py:38)
REGISTRIES_MANAGER_ADDRESS_LOCAL  # unused variable (autonomy/chain/constants.py:39)
SERVICE_MANAGER_ADDRESS_LOCAL  # unused variable (autonomy/chain/constants.py:40)
SERVICE_REGISTRY_ADDRESS_LOCAL  # unused variable (autonomy/chain/constants.py:43)
COMPONENT_REGISTRY_ADDRESS_ETHEREUM  # unused variable (autonomy/chain/constants.py:45)
AGENT_REGISTRY_ADDRESS_ETHEREUM  # unused variable (autonomy/chain/constants.py:46)
REGISTRIES_MANAGER_ADDRESS_ETHEREUM  # unused variable (autonomy/chain/constants.py:47)
SERVICE_MANAGER_ADDRESS_ETHEREUM  # unused variable (autonomy/chain/constants.py:48)
SERVICE_REGISTRY_ADDRESS_ETHEREUM  # unused variable (autonomy/chain/constants.py:49)
COMPONENT_REGISTRY_ADDRESS_GOERLI  # unused variable (autonomy/chain/constants.py:52)
AGENT_REGISTRY_ADDRESS_GOERLI  # unused variable (autonomy/chain/constants.py:53)
REGISTRIES_MANAGER_ADDRESS_GOERLI  # unused variable (autonomy/chain/constants.py:54)
SERVICE_MANAGER_ADDRESS_GOERLI  # unused variable (autonomy/chain/constants.py:55)
SERVICE_REGISTRY_ADDRESS_GOERLI  # unused variable (autonomy/chain/constants.py:56)
COMPONENT_REGISTRY_ADDRESS_CUSTOM  # unused variable (autonomy/chain/constants.py:59)
AGENT_REGISTRY_ADDRESS_CUSTOM  # unused variable (autonomy/chain/constants.py:60)
REGISTRIES_MANAGER_ADDRESS_CUSTOM  # unused variable (autonomy/chain/constants.py:61)
SERVICE_MANAGER_ADDRESS_CUSTOM  # unused variable (autonomy/chain/constants.py:62)
SERVICE_REGISTRY_ADDRESS_CUSTOM  # unused variable (autonomy/chain/constants.py:63)
abi_file  # unused variable (open-autonomy/autonomy/chain/config.py:65)
AGENT_REGISTRY_CONTRACT  # unused variable (open-autonomy/autonomy/chain/constants.py:72)
SERVICE_MANAGER_CONTRACT  # unused variable (open-autonomy/autonomy/chain/constants.py:74)
SERVICE_REGISTRY_CONTRACT  # unused variable (open-autonomy/autonomy/chain/constants.py:75)
ComponentRegistryContract  # unused class (open-autonomy/autonomy/data/packages/valory/contracts/component_registry/contract.py:39)
contract_id  # unused variable (open-autonomy/autonomy/data/packages/valory/contracts/component_registry/contract.py:42)
get_raw_transaction  # unused method (open-autonomy/autonomy/data/packages/valory/contracts/component_registry/contract.py:50)
kwargs  # unused variable (open-autonomy/autonomy/data/packages/valory/contracts/component_registry/contract.py:52)
get_raw_message  # unused method (open-autonomy/autonomy/data/packages/valory/contracts/component_registry/contract.py:57)
kwargs  # unused variable (open-autonomy/autonomy/data/packages/valory/contracts/component_registry/contract.py:59)
get_state  # unused method (open-autonomy/autonomy/data/packages/valory/contracts/component_registry/contract.py:64)
kwargs  # unused variable (open-autonomy/autonomy/data/packages/valory/contracts/component_registry/contract.py:66)
RegistriesManagerContract  # unused class (open-autonomy/autonomy/data/packages/valory/contracts/registries_manager/contract.py:39)
contract_id  # unused variable (open-autonomy/autonomy/data/packages/valory/contracts/registries_manager/contract.py:42)
get_raw_transaction  # unused method (open-autonomy/autonomy/data/packages/valory/contracts/registries_manager/contract.py:50)
kwargs  # unused variable (open-autonomy/autonomy/data/packages/valory/contracts/registries_manager/contract.py:52)
get_raw_message  # unused method (open-autonomy/autonomy/data/packages/valory/contracts/registries_manager/contract.py:57)
kwargs  # unused variable (open-autonomy/autonomy/data/packages/valory/contracts/registries_manager/contract.py:59)
get_state  # unused method (open-autonomy/autonomy/data/packages/valory/contracts/registries_manager/contract.py:64)
kwargs  # unused variable (open-autonomy/autonomy/data/packages/valory/contracts/registries_manager/contract.py:66)
path_to_contract  # unused variable (open-autonomy/autonomy/data/contracts/service_registry/tests/test_contract.py:54)
ledger_identifier  # unused variable (open-autonomy/autonomy/data/contracts/service_registry/tests/test_contract.py:55)
finish_contract_deployment  # unused method (open-autonomy/autonomy/data/contracts/service_registry/tests/test_contract.py:57)
_deploy_contract  # unused method (open-autonomy/autonomy/data/contracts/service_registry/tests/test_contract.py:62)
deployer_crypto  # unused variable (open-autonomy/autonomy/data/contracts/service_registry/tests/test_contract.py:67)
gas  # unused variable (open-autonomy/autonomy/data/contracts/service_registry/tests/test_contract.py:68)
connection  # unused function (open-autonomy/autonomy/cli/mint.py:127)
skill  # unused function (open-autonomy/autonomy/cli/mint.py:155)
AgentRegistryContract  # unused class (open-autonomy/autonomy/data/contracts/agent_registry/contract.py:38)
default_account  # unused attribute (autonomy/cli/helpers/chain.py:85:)
activate  # unused function (autonomy/cli/service.py:48)
register  # unused function (autonomy/cli/service.py:78)
deploy  # unused function (autonomy/cli/service.py:126)
DEPLOYED  # unused variable (open-autonomy/autonomy/chain/base.py:53)
TERMINATED_BONDED  # unused variable (open-autonomy/autonomy/chain/base.py:54)
contract  # unused function (open-autonomy/autonomy/cli/mint.py:106)
directory  # unused attribute (open-autonomy/autonomy/cli/helpers/analyse.py:43)
_check_dialogues  # unused function (open-autonomy/autonomy/cli/analyse.py:216)
parse_file  # unused function (open-autonomy/autonomy/analyse/abci/logs.py:53)
n_agents  # unused attribute (open-autonomy/autonomy/analyse/logs/collection.py:48)
insert_one  # unused method (open-autonomy/autonomy/analyse/logs/db.py:143)
_parse_logs  # unused function (open-autonomy/autonomy/cli/analyse.py:165)
_check_service  # unused function (open-autonomy/autonomy/cli/analyse.py:289)
get_metavar  # unused method (autonomy/cli/utils/click_utils.py:124)
_terminate  # unused function (autonomy/cli/service.py:162)
_unbond  # unused function (autonomy/cli/service.py:186)
_info  # unused function (autonomy/cli/service.py:210)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the data classes for the reset_pause_abci application."""

from enum import Enum
from typing import Dict, Optional, Set, Tuple

from packages.valory.skills.abstract_round_abci.base import (
    AbciApp,
    AbciAppTransitionFunction,
    AppState,
    BaseSynchronizedData,
    CollectSameUntilThresholdRound,
    DegenerateRound,
)
from packages.valory.skills.reset_pause_abci.payloads import ResetPausePayload


class Event(Enum):
    """Event enumeration for the reset_pause_abci app."""

    DONE = "done"
    ROUND_TIMEOUT = "round_timeout"
    NO_MAJORITY = "no_majority"
    RESET_AND_PAUSE_TIMEOUT = "reset_and_pause_timeout"


class ResetAndPauseRound(CollectSameUntilThresholdRound):
    """A round that represents that consensus is reached (the final round)"""

    payload_class = ResetPausePayload
    _allow_rejoin_payloads = True
    synchronized_data_class = BaseSynchronizedData

    def end_block(self) -> Optional[Tuple[BaseSynchronizedData, Event]]:
        """Process the end of the block."""
        if self.threshold_reached:
            return self.synchronized_data.create(), Event.DONE
        if not self.is_majority_possible(
            self.collection, self.synchronized_data.nb_participants
        ):
            return self.synchronized_data, Event.NO_MAJORITY
        return None


class FinishedResetAndPauseRound(DegenerateRound):
    """A round that represents reset and pause has finished"""


class FinishedResetAndPauseErrorRound(DegenerateRound):
    """A round that represents reset and pause has finished with errors"""


class ResetPauseAbciApp(AbciApp[Event]):
    """ResetPauseAbciApp

    Initial round: ResetAndPauseRound

    Initial states: {ResetAndPauseRound}

    Transition states:
        0. ResetAndPauseRound
            - done: 1.
            - reset and pause timeout: 2.
            - no majority: 2.
        1. FinishedResetAndPauseRound
        2. FinishedResetAndPauseErrorRound

    Final states: {FinishedResetAndPauseErrorRound, FinishedResetAndPauseRound}

    Timeouts:
        round timeout: 30.0
        reset and pause timeout: 30.0
    """

    initial_round_cls: AppState = ResetAndPauseRound
    transition_function: AbciAppTransitionFunction = {
        ResetAndPauseRound: {
            Event.DONE: FinishedResetAndPauseRound,
            Event.RESET_AND_PAUSE_TIMEOUT: FinishedResetAndPauseErrorRound,
            Event.NO_MAJORITY: FinishedResetAndPauseErrorRound,
        },
        FinishedResetAndPauseRound: {},
        FinishedResetAndPauseErrorRound: {},
    }
    final_states: Set[AppState] = {
        FinishedResetAndPauseRound,
        FinishedResetAndPauseErrorRound,
    }
    event_to_timeout: Dict[Event, float] = {
        Event.ROUND_TIMEOUT: 30.0,
        Event.RESET_AND_PAUSE_TIMEOUT: 30.0,
    }
    db_pre_conditions: Dict[AppState, Set[str]] = {ResetAndPauseRound: set()}
    db_post_conditions: Dict[AppState, Set[str]] = {
        FinishedResetAndPauseRound: set(),
        FinishedResetAndPauseErrorRound: set(),
    }



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""FSM scaffold tools."""

import json
import re
from abc import ABC, abstractmethod
from pathlib import Path
from textwrap import dedent, indent
from typing import Dict, List

from aea.cli.utils.context import Context

from autonomy.analyse.abci.app_spec import DFA
from autonomy.fsm.scaffold.constants import (
    ABCI_APP,
    BASE_BEHAVIOUR,
    BEHAVIOUR,
    PAYLOAD,
    ROUND,
    ROUND_BEHAVIOUR,
    TEMPLATE_INDENTATION,
)


def _remove_quotes(input_str: str) -> str:
    """Remove single or double quotes from a string."""
    return input_str.replace("'", "").replace('"', "")


def _indent_wrapper(lines: str) -> str:
    """Indentation"""
    return indent(lines, TEMPLATE_INDENTATION).strip()


class AbstractFileGenerator(ABC):
    """An abstract class for file generators."""

    FILENAME: str

    def __init__(self, ctx: Context, skill_name: str, dfa: DFA) -> None:
        """Initialize the abstract file generator."""
        self.ctx = ctx
        self.skill_name = skill_name
        self.dfa = dfa

    @abstractmethod
    def get_file_content(self) -> str:
        """Get file content."""

    def write_file(self, output_dir: Path) -> None:
        """Write the file to output_dir/FILENAME."""
        (output_dir / self.FILENAME).write_text(dedent(self.get_file_content()))

    @property
    def abci_app_name(self) -> str:
        """ABCI app class name"""
        return self.dfa.label.split(".")[-1]

    @property
    def fsm_name(self) -> str:
        """FSM base name"""
        return re.sub(ABCI_APP, "", self.abci_app_name, flags=re.IGNORECASE)

    @property
    def author(self) -> str:
        """Author"""
        return self.ctx.agent_config.author

    @property
    def all_rounds(self) -> List[str]:
        """Rounds"""
        return sorted(self.dfa.states)

    @property
    def degenerate_rounds(self) -> List[str]:
        """Degenerate rounds"""
        return sorted(self.dfa.final_states)

    @property
    def rounds(self) -> List[str]:
        """Non-degenerate rounds"""
        return sorted(self.dfa.states - self.dfa.final_states)

    @property
    def behaviours(self) -> List[str]:
        """Behaviours"""
        return [s.replace(ROUND, BEHAVIOUR) for s in self.rounds]

    @property
    def payloads(self) -> List[str]:
        """Payloads"""
        return [s.replace(ROUND, PAYLOAD) for s in self.rounds]

    @property  # TODO: functools cached property
    def template_kwargs(self) -> Dict[str, str]:
        """All keywords for string formatting of templates"""

        events_list = [
            f'{event_name} = "{event_name.lower()}"'
            for event_name in self.dfa.alphabet_in
        ]

        tf = json.dumps(self.dfa.parse_transition_func(), indent=4)
        behaviours = json.dumps(self.behaviours, indent=4)

        return dict(
            author=self.author,
            skill_name=self.skill_name,
            FSMName=self.fsm_name,
            AbciApp=self.abci_app_name,
            rounds=_indent_wrapper(",\n".join(self.rounds)),
            all_rounds=_indent_wrapper(",\n".join(self.all_rounds)),
            behaviours=_indent_wrapper(",\n".join(self.behaviours)),
            payloads=_indent_wrapper(",\n".join(self.payloads)),
            events=_indent_wrapper("\n".join(events_list)),
            initial_round_cls=self.dfa.default_start_state,
            initial_states=_remove_quotes(str(self.dfa.start_states)),
            transition_function=_indent_wrapper(_remove_quotes(str(tf))),
            final_states=_remove_quotes(str(self.dfa.final_states)),
            BaseBehaviourCls=re.sub(
                ABCI_APP, BASE_BEHAVIOUR, self.abci_app_name, flags=re.IGNORECASE
            ),
            RoundBehaviourCls=re.sub(
                ABCI_APP, ROUND_BEHAVIOUR, self.abci_app_name, flags=re.IGNORECASE
            ),
            InitialBehaviourCls=self.dfa.default_start_state.replace(ROUND, BEHAVIOUR),
            round_behaviours=_indent_wrapper(_remove_quotes(str(behaviours))),
            db_pre_conditions=_indent_wrapper(
                "\n".join([f"\t{round}: []," for round in self.dfa.start_states])
            ),
            db_post_conditions=_indent_wrapper(
                "\n".join([f"\t{round}: []," for round in self.dfa.final_states])
            ),
        )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
# pylint: disable=unused-import

"""Integration tests for the valory/register_termination skill."""
from pathlib import Path

import pytest
from aea_test_autonomy.base_test_classes.agents import BaseTestEnd2EndExecution
from aea_test_autonomy.fixture_helpers import (  # noqa: F401
    UseRegistries,
    abci_host,
    abci_port,
    flask_tendermint,
    hardhat_addr,
    hardhat_port,
    ipfs_daemon,
    ipfs_domain,
    key_pairs,
    registries_scope_class,
    tendermint_port,
)


TARGET_AGENT = "valory/test_ipfs:0.1.0"
TARGET_SKILL = "valory/test_ipfs_abci:0.1.0"
TIME_TO_FINISH = 60  # 1 minute


@pytest.mark.e2e
@pytest.mark.integration
@pytest.mark.parametrize("nb_nodes", (1,))
class TestIpfs(
    BaseTestEnd2EndExecution,
):
    """Test whether uploading and receiving from IPFS works.."""

    package_registry_src_rel = Path(__file__).parents[4]
    agent_package = TARGET_AGENT
    skill_package = TARGET_SKILL
    wait_to_finish = TIME_TO_FINISH
    strict_check_strings = (
        "Single object uploading & downloading works.",
        "Multiple object uploading & downloading works.",
    )
